# pylint: disable=too-many-arguments
import numpy as np
from ...utils import get_cuda_srcfile, updiv
from ...reconstruction.filtering import get_next_power
from ...reconstruction.filtering_cuda import CudaSinoFilter

# pylint: disable=too-many-arguments,too-many-instance-attributes,too-many-function-args


class HelicalSinoFilter(CudaSinoFilter):
    def __init__(
        self,
        sino_shape,
        filter_name=None,
        padding_mode="zeros",
        extra_options=None,
        cuda_options=None,
    ):
        """Derived from nabu.reconstruction.filtering.SinoFilter
        It is used by helical_chunked_regridded pipeline.
        The shape of the processed sino, as a matter of fact which is due to the helical_chunked_regridded.py module
        which is using the here present class, is always, but not necessarily [nangles, nslices, nhorizontal]
        with nslices = 1. This  because  helical_chunked_regridded.py after a first preprocessing phase,
        always processes slices one by one.
        In helical_chunked_regridded .py, the call to the filter_sino method here contained  is
        followed by  the weight redistribution ( done by another module), which solves the HA problem,
        and the backprojection. The latter is performed by fbp.py or hbp.py

        The reason for having this class, derived from nabu.reconstruction.filtering.SinoFilter,
        is that the padding mechanism here implemented incorporates the padding with the available
        theta+180 projection on the half-tomo side.

        """
        super().__init__(
            sino_shape,
            filter_name=filter_name,
            padding_mode=padding_mode,
            extra_options=extra_options,
            cuda_options=cuda_options,
        )

        self._init_pad_kernel()

    def _check_array(self, arr):
        """
        This class may work with an arbitrary number of projections. This is a consequence
        of the first implementation of the helical pipeline. In the first implementation
        the slices were reconstructed by backprojecting several turns, and the number of useful projections
        was different from the beginning or end, to the center of the scan.
        Now in helical_chunked_regridded.py the number of projections is fixed. The only relic left
        is that the present class may work with an arbitrary number of projections.
        """
        if arr.dtype != np.float32:
            raise ValueError("Expected data type = numpy.float32")
        if arr.shape[1:] != self.sino_shape[1:]:
            raise ValueError("Expected sinogram shape %s, got %s" % (self.sino_shape, arr.shape))

    def _init_pad_kernel(self):
        """The four possible padding kernels.
        The first two, compared to nabu.reconstruction.filtering.SinoFilter
        can work with an arbitrary number of projection.
        The latter two implement the padding with the available information from theta+180.
        """

        self.kern_args = (self.d_sino_f, self.d_filter_f)
        self.kern_args += self.d_sino_f.shape[::-1]
        self._pad_mirror_edges_kernel = self.cuda.kernel(
            "padding",
            filename=get_cuda_srcfile("helical_padding.cu"),
            signature="PPfiiiii",
            options=[str("-DMIRROR_EDGES")],
        )
        self._pad_mirror_constant_kernel = self.cuda.kernel(
            "padding",
            filename=get_cuda_srcfile("helical_padding.cu"),
            signature="PPfiiiiiff",
            options=[str("-DMIRROR_CONSTANT")],
        )

        self._pad_mirror_edges_variable_rot_pos_kernel = self.cuda.kernel(
            "padding",
            filename=get_cuda_srcfile("helical_padding.cu"),
            signature="PPPiiiii",
            options=[str("-DMIRROR_EDGES_VARIABLE_ROT_POS")],
        )
        self._pad_mirror_constant_variable_rot_pos_kernel = self.cuda.kernel(
            "padding",
            filename=get_cuda_srcfile("helical_padding.cu"),
            signature="PPPiiiiiff",
            options=[str("-DMIRROR_CONSTANT_VARIABLE_ROT_POS")],
        )

        self.d_mirror_indexes = self.cuda.allocate_array(
            "d_mirror_indexes", (self.sino_padded_shape[-2],), dtype=np.int32
        )
        self.d_variable_rot_pos = self.cuda.allocate_array(
            "d_variable_rot_pos", (self.sino_padded_shape[-2],), dtype=np.float32
        )
        self._pad_edges_kernel = self.cuda.kernel(
            "padding_edge", filename=get_cuda_srcfile("padding.cu"), signature="Piiiiiiii"
        )
        self._pad_block = (32, 32, 1)
        self._pad_grid = tuple([updiv(n, p) for n, p in zip(self.sino_padded_shape[::-1], self._pad_block)])

    def _pad_sino(self, sino, mirror_indexes=None, rot_center=None):
        """redefined here to adapt the memory copy to the lenght of the sino argument
        which, in the general helical case may be varying
        """
        if mirror_indexes is None:
            self._pad_sino_simple(sino)

        self.d_mirror_indexes[:] = np.zeros([len(self.d_mirror_indexes)], np.int32)
        self.d_mirror_indexes[: len(mirror_indexes)] = mirror_indexes.astype(np.int32)

        if np.isscalar(rot_center):
            argument_rot_center = rot_center
            tmp_pad_mirror_edges_kernel = self._pad_mirror_edges_kernel
            tmp_pad_mirror_constant_kernel = self._pad_mirror_constant_kernel
        else:
            self.d_variable_rot_pos[: len(rot_center)] = rot_center
            argument_rot_center = self.d_variable_rot_pos
            tmp_pad_mirror_edges_kernel = self._pad_mirror_edges_variable_rot_pos_kernel
            tmp_pad_mirror_constant_kernel = self._pad_mirror_constant_variable_rot_pos_kernel

        self.d_sino_padded[: len(sino), : self.dwidth] = sino[:]
        if self.padding_mode == "edges":
            tmp_pad_mirror_edges_kernel(
                self.d_sino_padded,
                self.d_mirror_indexes,
                argument_rot_center,
                self.dwidth,
                self.n_angles,
                self.dwidth_padded,
                self.pad_left,
                self.pad_right,
                grid=self._pad_grid,
                block=self._pad_block,
            )
        else:
            tmp_pad_mirror_constant_kernel(
                self.d_sino_padded,
                self.d_mirror_indexes,
                argument_rot_center,
                self.dwidth,
                self.n_angles,
                self.dwidth_padded,
                self.pad_left,
                self.pad_right,
                0.0,
                0.0,
                grid=self._pad_grid,
                block=self._pad_block,
            )

    def _pad_sino_simple(self, sino):
        if self.padding_mode == "edges":
            self.d_sino_padded[: len(sino), : self.dwidth] = sino[:]
            self._pad_edges_kernel(
                self.d_sino_padded,
                self.dwidth,
                self.n_angles,
                self.dwidth_padded,
                self.n_angles,
                self.pad_left,
                self.pad_right,
                0,
                0,
                grid=self._pad_grid,
                block=self._pad_block,
            )
        else:  # zeros
            self.d_sino_padded.fill(0)
            if self.ndim == 2:
                self.d_sino_padded[: len(sino), : self.dwidth] = sino[:]
            else:
                self.d_sino_padded[: len(sino), :, : self.dwidth] = sino[:]

    def filter_sino(self, sino, mirror_indexes=None, rot_center=None, output=None, no_output=False):
        """
        Perform the sinogram siltering.
        redefined here to use also mirror data
        Parameters
        ----------
        sino: numpy.ndarray or pycuda.gpuarray.GPUArray
            Input sinogram (2D or 3D)
        output: numpy.ndarray or pycuda.gpuarray.GPUArray, optional
            Output array.
        no_output: bool, optional
            If set to True, no copy is be done. The resulting data lies
            in self.d_sino_padded.
        """
        self._check_array(sino)
        # copy2d/copy3d
        self._pad_sino(sino, mirror_indexes=mirror_indexes, rot_center=rot_center)
        # FFT
        self.fft.fft(self.d_sino_padded, output=self.d_sino_f)

        # multiply padded sinogram with filter in the Fourier domain
        self.mult_kernel(*self.kern_args)  # TODO tune block size ?

        # iFFT
        self.fft.ifft(self.d_sino_f, output=self.d_sino_padded)

        # return
        if no_output:
            return self.d_sino_padded
        if output is None:
            res = np.zeros(self.sino_shape, dtype=np.float32)
            # can't do memcpy2d D->H ? (self.d_sino_padded[:, w]) I have to get()
            sino_ref = self.d_sino_padded.get()
        else:
            res = output
            sino_ref = self.d_sino_padded
        if self.ndim == 2:
            res[:] = sino_ref[:, : self.dwidth]
        else:
            res[:] = sino_ref[:, :, : self.dwidth]
        return res

    def _calculate_shapes(self, sino_shape):
        self.ndim = len(sino_shape)
        if self.ndim == 2:
            n_angles, dwidth = sino_shape
            n_sinos = 1
        elif self.ndim == 3:
            n_sinos, n_angles, dwidth = sino_shape
        else:
            raise ValueError("Invalid sinogram number of dimensions")
        self.sino_shape = sino_shape
        self.n_angles = n_angles
        self.dwidth = dwidth
        # int() is crucial here ! Otherwise some pycuda arguments (ex. memcpy2D)
        # will not work with numpy.int64 (as for 2018.X)

        ### the original get_next_power used in  nabu gives a lower ram footprint
        self.dwidth_padded = 2 * int(get_next_power(self.dwidth))

        self.sino_padded_shape = (n_angles, self.dwidth_padded)
        if self.ndim == 3:
            self.sino_padded_shape = (n_sinos,) + self.sino_padded_shape
        sino_f_shape = list(self.sino_padded_shape)
        sino_f_shape[-1] = sino_f_shape[-1] // 2 + 1
        self.sino_f_shape = tuple(sino_f_shape)
        #
        self.pad_left = (self.dwidth_padded - self.dwidth) // 2
        self.pad_right = self.dwidth_padded - self.dwidth - self.pad_left



from multiprocessing import get_context
from multiprocessing.pool import Pool
from .fft_base import _BaseVKFFT
from ..opencl.processing import OpenCLProcessing

try:
    from pyvkfft.opencl import VkFFTApp as vk_clfft

    __has_vkfft__ = True
except (ImportError, OSError):
    __has_vkfft__ = False
    vk_clfft = None


class VKCLFFT(_BaseVKFFT):
    """
    OpenCL FFT, using VKFFT backend
    """

    implem = "vkfft"
    backend = "opencl"
    ProcessingCls = OpenCLProcessing
    vkffs_cls = vk_clfft

    def _init_backend(self, backend_options):
        super()._init_backend(backend_options)
        self._vkfft_other_init_kwargs = {"queue": self.processing.queue}


def _has_vkfft(x):
    # should be run from within a Process
    try:
        from nabu.processing.fft_opencl import VKCLFFT, __has_vkfft__

        if not __has_vkfft__:
            return False
        vk = VKCLFFT((16,), "f")
        avail = True
    except (RuntimeError, OSError):
        avail = False
    return avail


def has_vkfft(safe=True):
    """
    Determine whether pyvkfft is available.
    This function cannot be tested from a notebook/console, a proper entry point has to be created (if __name__ == "__main__").
    """
    if not safe:
        return _has_vkfft(None)
    ctx = get_context("spawn")
    with Pool(1, context=ctx) as p:
        v = p.map(_has_vkfft, [1])[0]
    return v



from distutils.version import StrictVersion
from typing import Optional, Union
import logging
import functools
import numpy
from tomoscan.scanbase import TomoScanBase
from tomoscan.volumebase import VolumeBase
from nxtomo.utils.transformation import build_matrix, DetYFlipTransformation
from silx.utils.enum import Enum as _Enum
from scipy.fft import rfftn as local_fftn
from scipy.fft import irfftn as local_ifftn
from ..overlap import OverlapStitchingStrategy, ImageStichOverlapKernel
from ..alignment import AlignmentAxis1, AlignmentAxis2, PaddedRawData
from ...misc import fourier_filters
from ...estimation.alignment import AlignmentBase
from ...resources.dataset_analyzer import HDF5DatasetAnalyzer
from ...resources.nxflatfield import update_dataset_info_flats_darks

try:
    import itk
except ImportError:
    has_itk = False
else:
    has_itk = True

_logger = logging.getLogger(__name__)


try:
    from skimage.registration import phase_cross_correlation
except ImportError:
    _logger.warning(
        "Unable to load skimage. Please install it if you want to use it for finding shifts from `find_relative_shifts`"
    )
    __has_sk_phase_correlation__ = False
else:
    __has_sk_phase_correlation__ = True


class ShiftAlgorithm(_Enum):
    """All generic shift search algorithm"""

    NABU_FFT = "nabu-fft"
    SKIMAGE = "skimage"
    ITK_IMG_REG_V4 = "itk-img-reg-v4"
    NONE = "None"

    # In the case of shift search on radio along axis 2 (or axis x in image space) we can benefit from the existing
    # nabu algorithm such as growing-window or sliding-window
    CENTERED = "centered"
    GLOBAL = "global"
    SLIDING_WINDOW = "sliding-window"
    GROWING_WINDOW = "growing-window"
    SINO_COARSE_TO_FINE = "sino-coarse-to-fine"
    COMPOSITE_COARSE_TO_FINE = "composite-coarse-to-fine"

    @classmethod
    def from_value(cls, value):
        if value in ("", None):
            return ShiftAlgorithm.NONE
        else:
            return super().from_value(value=value)


def find_frame_relative_shifts(
    overlap_upper_frame: numpy.ndarray,
    overlap_lower_frame: numpy.ndarray,
    estimated_shifts: tuple,
    overlap_axis: int,
    x_cross_correlation_function=None,
    y_cross_correlation_function=None,
    x_shifts_params: Optional[dict] = None,
    y_shifts_params: Optional[dict] = None,
):
    """
    :param overlap_axis: axis in [0, 1] on which the overlap exists. In image space. So 0 is aka y and 1 as x
    """
    if not overlap_axis in (0, 1):
        raise ValueError(f"overlap_axis should be in (0, 1). Get {overlap_axis}")
    from nabu.stitching.config import (
        KEY_LOW_PASS_FILTER,
        KEY_HIGH_PASS_FILTER,
    )  # avoid cyclic import

    x_cross_correlation_function = ShiftAlgorithm.from_value(x_cross_correlation_function)
    y_cross_correlation_function = ShiftAlgorithm.from_value(y_cross_correlation_function)

    if x_shifts_params is None:
        x_shifts_params = {}

    if y_shifts_params is None:
        y_shifts_params = {}

    # apply filtering if any
    def _str_to_int(value):
        if isinstance(value, str):
            value = value.lstrip("'").lstrip('"')
            value = value.rstrip("'").rstrip('"')
            value = int(value)
        return value

    low_pass = _str_to_int(x_shifts_params.get(KEY_LOW_PASS_FILTER, y_shifts_params.get(KEY_LOW_PASS_FILTER, None)))
    high_pass = _str_to_int(x_shifts_params.get(KEY_HIGH_PASS_FILTER, y_shifts_params.get(KEY_HIGH_PASS_FILTER, None)))
    if high_pass is None and low_pass is None:
        pass
    else:
        if low_pass is None:
            low_pass = 1
        if high_pass is None:
            high_pass = 20
        _logger.info(f"filter image for shift search (low_pass={low_pass}, high_pass={high_pass})")
        img_filter = fourier_filters.get_bandpass_filter(
            overlap_upper_frame.shape[-2:],
            cutoff_lowpass=low_pass,
            cutoff_highpass=high_pass,
            use_rfft=True,
            data_type=overlap_upper_frame.dtype,
        )
        overlap_upper_frame = local_ifftn(
            local_fftn(overlap_upper_frame, axes=(-2, -1)) * img_filter, axes=(-2, -1)
        ).real
        overlap_lower_frame = local_ifftn(
            local_fftn(overlap_lower_frame, axes=(-2, -1)) * img_filter, axes=(-2, -1)
        ).real

    # compute shifts
    initial_shifts = numpy.array(estimated_shifts).copy()
    extra_shifts = numpy.array([0.0, 0.0])

    def skimage_proxy(img1, img2):
        if not __has_sk_phase_correlation__:
            raise ValueError("scikit-image not installed. Cannot do phase correlation from it")
        else:
            found_shift, _, _ = phase_cross_correlation(reference_image=img1, moving_image=img2, space="real")
            return -found_shift

    shift_methods = {
        ShiftAlgorithm.NABU_FFT: functools.partial(
            find_shift_correlate, img1=overlap_upper_frame, img2=overlap_lower_frame
        ),
        ShiftAlgorithm.SKIMAGE: functools.partial(skimage_proxy, img1=overlap_upper_frame, img2=overlap_lower_frame),
        ShiftAlgorithm.ITK_IMG_REG_V4: functools.partial(
            find_shift_with_itk, img1=overlap_upper_frame, img2=overlap_lower_frame
        ),
        ShiftAlgorithm.NONE: functools.partial(lambda: (0.0, 0.0)),
    }

    res_algo = {}
    for shift_alg in set((x_cross_correlation_function, y_cross_correlation_function)):
        if shift_alg not in shift_methods:
            raise ValueError(f"requested image alignment function not handled ({shift_alg})")
        try:
            res_algo[shift_alg] = shift_methods[shift_alg]()
        except Exception as e:
            _logger.error(f"Failed to find shift from {shift_alg.value}. Error is {e}")
            res_algo[shift_alg] = (0, 0)

    extra_shifts = (
        res_algo[y_cross_correlation_function][0],
        res_algo[x_cross_correlation_function][1],
    )

    final_rel_shifts = numpy.array(extra_shifts) + initial_shifts
    return tuple([int(shift) for shift in final_rel_shifts])


def find_volumes_relative_shifts(
    upper_volume: VolumeBase,
    lower_volume: VolumeBase,
    overlap_axis: int,
    estimated_shifts,
    dim_axis_1: int,
    dtype,
    flip_ud_upper_frame: bool = False,
    flip_ud_lower_frame: bool = False,
    slice_for_shift: Union[int, str] = "middle",
    x_cross_correlation_function=None,
    y_cross_correlation_function=None,
    x_shifts_params: Optional[dict] = None,
    y_shifts_params: Optional[dict] = None,
    alignment_axis_2="center",
    alignment_axis_1="center",
):
    """

    :param int dim_axis_1: axis 1 dimension (to handle axis 1 alignment)
    """
    if y_shifts_params is None:
        y_shifts_params = {}

    if x_shifts_params is None:
        x_shifts_params = {}
    # convert from overlap_axis (3D acquisition space) to overlap_axis_proj_space.
    if overlap_axis == 1:
        raise NotImplementedError("finding projection shift along axis 1 is not handled for projections")
    elif overlap_axis == 0:
        overlap_axis_proj_space = 0
    elif overlap_axis == 2:
        overlap_axis_proj_space = 1
    else:
        raise ValueError(f"Stitching is done in 3D space. Expect axis to be in [0,2]. Get {overlap_axis}")

    alignment_axis_2 = AlignmentAxis2.from_value(alignment_axis_2)
    alignment_axis_1 = AlignmentAxis1.from_value(alignment_axis_1)
    assert dim_axis_1 > 0, "dim_axis_1 <= 0"

    if isinstance(slice_for_shift, str):
        if slice_for_shift == "first":
            slice_for_shift = 0
        elif slice_for_shift == "last":
            slice_for_shift = dim_axis_1
        elif slice_for_shift == "middle":
            slice_for_shift = dim_axis_1 // 2
        else:
            raise ValueError("invalid slice provided to search shift", slice_for_shift)

    def get_slice_along_axis_1(volume: VolumeBase, index: int):
        assert isinstance(index, int), f"index should be an int, {type(index)} provided"
        volume_shape = volume.get_volume_shape()
        if alignment_axis_1 is AlignmentAxis1.BACK:
            front_empty_width = dim_axis_1 - volume_shape[1]
            if index < front_empty_width:
                return PaddedRawData.get_empty_frame(shape=(volume_shape[0], volume_shape[2]), dtype=dtype)
            else:
                return volume.get_slice(index=index - front_empty_width, axis=1)
        elif alignment_axis_1 is AlignmentAxis1.FRONT:
            if index >= volume_shape[1]:
                return PaddedRawData.get_empty_frame(shape=(volume_shape[0], volume_shape[2]), dtype=dtype)
            else:
                return volume.get_slice(index=index, axis=1)
        elif alignment_axis_1 is AlignmentAxis1.CENTER:
            front_empty_width = (dim_axis_1 - volume_shape[1]) // 2
            back_empty_width = dim_axis_1 - front_empty_width
            if index < front_empty_width or index > back_empty_width:
                return PaddedRawData.get_empty_frame(shape=(volume_shape[0], volume_shape[2]), dtype=dtype)
            else:
                return volume.get_slice(index=index - front_empty_width, axis=1)
        else:
            raise TypeError(f"unmanaged alignment mode {alignment_axis_1.value}")

    upper_frame = get_slice_along_axis_1(upper_volume, index=slice_for_shift)
    lower_frame = get_slice_along_axis_1(lower_volume, index=slice_for_shift)
    if flip_ud_upper_frame:
        upper_frame = numpy.flipud(upper_frame.copy())
    if flip_ud_lower_frame:
        lower_frame = numpy.flipud(lower_frame.copy())

    from nabu.stitching.config import KEY_WINDOW_SIZE  # avoid cyclic import

    w_window_size = int(y_shifts_params.get(KEY_WINDOW_SIZE, 400))
    start_overlap = max(estimated_shifts[0] // 2 - w_window_size // 2, 0)
    end_overlap = min(estimated_shifts[0] // 2 + w_window_size // 2, min(upper_frame.shape[0], lower_frame.shape[0]))

    if start_overlap == 0:
        overlap_upper_frame = upper_frame[-end_overlap:]
    else:
        overlap_upper_frame = upper_frame[-end_overlap:-start_overlap]
    overlap_lower_frame = lower_frame[start_overlap:end_overlap]

    # align if necessary
    if overlap_upper_frame.shape[1] != overlap_lower_frame.shape[1]:
        overlap_frame_width = min(overlap_upper_frame.shape[1], overlap_lower_frame.shape[1])
        if alignment_axis_2 is AlignmentAxis2.CENTER:
            upper_frame_left_pos = overlap_upper_frame.shape[1] // 2 - overlap_frame_width // 2
            upper_frame_right_pos = upper_frame_left_pos + overlap_frame_width
            overlap_upper_frame = overlap_upper_frame[:, upper_frame_left_pos:upper_frame_right_pos]

            lower_frame_left_pos = overlap_lower_frame.shape[1] // 2 - overlap_frame_width // 2
            lower_frame_right_pos = lower_frame_left_pos + overlap_frame_width
            overlap_lower_frame = overlap_lower_frame[:, lower_frame_left_pos:lower_frame_right_pos]
        elif alignment_axis_2 is AlignmentAxis2.LEFT:
            overlap_upper_frame = overlap_upper_frame[:, :overlap_frame_width]
            overlap_lower_frame = overlap_lower_frame[:, :overlap_frame_width]
        elif alignment_axis_2 is AlignmentAxis2.RIGTH:
            overlap_upper_frame = overlap_upper_frame[:, -overlap_frame_width:]
            overlap_lower_frame = overlap_lower_frame[:, -overlap_frame_width:]
        else:
            raise ValueError(f"Alignement {alignment_axis_2.value} is not handled")

    if not overlap_upper_frame.shape == overlap_lower_frame.shape:
        raise ValueError(f"Fail to get consistant overlap ({overlap_upper_frame.shape} vs {overlap_lower_frame.shape})")

    return find_frame_relative_shifts(
        overlap_upper_frame=overlap_upper_frame,
        overlap_lower_frame=overlap_lower_frame,
        estimated_shifts=estimated_shifts,
        x_cross_correlation_function=x_cross_correlation_function,
        y_cross_correlation_function=y_cross_correlation_function,
        x_shifts_params=x_shifts_params,
        y_shifts_params=y_shifts_params,
        overlap_axis=overlap_axis_proj_space,
    )


from nabu.pipeline.estimators import estimate_cor


def find_projections_relative_shifts(
    upper_scan: TomoScanBase,
    lower_scan: TomoScanBase,
    estimated_shifts: tuple,
    axis: int,
    flip_ud_upper_frame: bool = False,
    flip_ud_lower_frame: bool = False,
    projection_for_shift: Union[int, str] = "middle",
    invert_order: bool = False,
    x_cross_correlation_function=None,
    y_cross_correlation_function=None,
    x_shifts_params: Optional[dict] = None,
    y_shifts_params: Optional[dict] = None,
) -> tuple:
    """
    deduce the relative shift between the two scans.
    Expected behavior:
    * compute expected overlap area from z_translations and (sample) pixel size
    * call an (optional) cross correlation function from the overlap area to compute the x shift and polish the y shift from `projection_for_shift`

    :param TomoScanBase scan_0:
    :param TomoScanBase scan_1:
    :param tuple estimated_shifts: 'a priori' shift estimation
    :param int axis: axis on which the overlap / stitching is happening. In the 3D space (sample, detector referential)
    :param bool flip_ud_upper_frame: is the upper frame flipped
    :param bool flip_ud_lower_frame: is the lower frame flipped
    :param Union[int,str] projection_for_shift: index fo the projection to use (in projection space or in scan space ?. For now in projection) or str. If str must be in (`middle`, `first`, `last`)
    :param bool invert_order: are projections inverted between the two scans (case if rotation angle are inverted)
    :param str x_cross_correlation_function: optional method to refine x shift from computing cross correlation. For now valid values are: ("skimage", "nabu-fft")
    :param str y_cross_correlation_function: optional method to refine y shift from computing cross correlation. For now valid values are: ("skimage", "nabu-fft")
    :param x_shifts_params: parameters to find the shift over x
    :param y_shifts_params: parameters to find the shift over y
    :return: relative shift of scan_1 with scan_0 as reference: (y_shift, x_shift)
    :rtype: tuple

    :warning: this function will flip left-right and up-down frames by default. So it will return shift according to this information
    """
    if x_shifts_params is None:
        x_shifts_params = {}
    if y_shifts_params is None:
        y_shifts_params = {}

    # convert from overlap_axis (3D acquisition space) to overlap_axis_proj_space.
    if axis == 1:
        axis_proj_space = 1
    elif axis == 0:
        axis_proj_space = 0
    elif axis == 2:
        raise NotImplementedError(
            "finding projection shift along axis 1 (x-ray direction) is not handled for projections"
        )
    else:
        raise ValueError(f"Stitching is done in 3D space. Expect axis to be in [0,2]. Get {axis}")

    x_cross_correlation_function = ShiftAlgorithm.from_value(x_cross_correlation_function)
    y_cross_correlation_function = ShiftAlgorithm.from_value(y_cross_correlation_function)

    # { handle specific use case (finding shift on scan) - when using nabu COR algorithms (for axis 2)
    if x_cross_correlation_function in (
        ShiftAlgorithm.SINO_COARSE_TO_FINE,
        ShiftAlgorithm.COMPOSITE_COARSE_TO_FINE,
        ShiftAlgorithm.CENTERED,
        ShiftAlgorithm.GLOBAL,
        ShiftAlgorithm.GROWING_WINDOW,
        ShiftAlgorithm.SLIDING_WINDOW,
    ):
        cor_options = x_shifts_params.copy()
        cor_options.pop("img_reg_method", None)
        # remove all none numeric options because estimate_cor will call 'literal_eval' on them

        upper_scan_dataset_info = HDF5DatasetAnalyzer(
            location=upper_scan.master_file, extra_options={"hdf5_entry": upper_scan.entry}
        )
        update_dataset_info_flats_darks(upper_scan_dataset_info, flatfield_mode=1)

        upper_scan_pos = estimate_cor(
            method=x_cross_correlation_function.value,
            dataset_info=upper_scan_dataset_info,
            cor_options=cor_options,
        )
        lower_scan_dataset_info = HDF5DatasetAnalyzer(
            location=lower_scan.master_file, extra_options={"hdf5_entry": lower_scan.entry}
        )
        update_dataset_info_flats_darks(lower_scan_dataset_info, flatfield_mode=1)
        lower_scan_pos = estimate_cor(
            method=x_cross_correlation_function.value,
            dataset_info=lower_scan_dataset_info,
            cor_options=cor_options,
        )

        estimated_shifts = tuple(
            [
                estimated_shifts[0],
                (lower_scan_pos - upper_scan_pos),
            ]
        )
        x_cross_correlation_function = ShiftAlgorithm.NONE

    # } else we will compute shift from the flat projections

    def get_flat_fielded_proj(
        scan: TomoScanBase, proj_index: int, reverse: bool, transformation_matrix: Optional[numpy.ndarray]
    ):
        first_proj_idx = sorted(lower_scan.projections.keys(), reverse=reverse)[proj_index]
        ff = scan.flat_field_correction(
            (scan.projections[first_proj_idx],),
            (first_proj_idx,),
        )[0]
        assert ff.ndim == 2, f"expects a single 2D frame. Get something with {ff.ndim} dimensions"
        if transformation_matrix is not None:
            assert (
                transformation_matrix.ndim == 2
            ), f"expects a 2D transformation matrix. Get a {transformation_matrix.ndim} D"
            if numpy.isclose(transformation_matrix[2, 2], -1):
                transformation_matrix[2, :] = 0
                transformation_matrix[0, 2] = 0
                transformation_matrix[2, 2] = 1
                ff = numpy.flipud(ff)
        return ff

    if isinstance(projection_for_shift, str):
        if projection_for_shift.lower() == "first":
            projection_for_shift = 0
        elif projection_for_shift.lower() == "last":
            projection_for_shift = -1
        elif projection_for_shift.lower() == "middle":
            projection_for_shift = len(upper_scan.projections) // 2
        else:
            try:
                projection_for_shift = int(projection_for_shift)
            except ValueError:
                raise ValueError(
                    f"{projection_for_shift} cannot be cast to an int and is not one of the possible ('first', 'last', 'middle')"
                )
    elif not isinstance(projection_for_shift, (int, numpy.number)):
        raise TypeError(
            f"projection_for_shift is expected to be an int. Not {type(projection_for_shift)} - {projection_for_shift}"
        )

    upper_scan_transformations = list(upper_scan.get_detector_transformations(tuple()))
    if flip_ud_upper_frame:
        upper_scan_transformations.append(DetYFlipTransformation(flip=True))
    upper_scan_trans_matrix = build_matrix(upper_scan_transformations)
    lower_scan_transformations = list(lower_scan.get_detector_transformations(tuple()))
    if flip_ud_lower_frame:
        lower_scan_transformations.append(DetYFlipTransformation(flip=True))
    lower_scan_trans_matrix = build_matrix(lower_scan_transformations)
    upper_proj = get_flat_fielded_proj(
        upper_scan,
        projection_for_shift,
        reverse=False,
        transformation_matrix=upper_scan_trans_matrix,
    )
    lower_proj = get_flat_fielded_proj(
        lower_scan,
        projection_for_shift,
        reverse=invert_order,
        transformation_matrix=lower_scan_trans_matrix,
    )

    from nabu.stitching.config import KEY_WINDOW_SIZE  # avoid cyclic import

    if axis_proj_space == 0:
        w_window_size = int(y_shifts_params.get(KEY_WINDOW_SIZE, 400))
    else:
        w_window_size = int(x_shifts_params.get(KEY_WINDOW_SIZE, 400))
    start_overlap = max(estimated_shifts[axis_proj_space] // 2 - w_window_size // 2, 0)
    end_overlap = min(
        estimated_shifts[axis_proj_space] // 2 + w_window_size // 2,
        min(upper_proj.shape[axis_proj_space], lower_proj.shape[axis_proj_space]),
    )
    o_upper_sel = numpy.array(range(-end_overlap, -start_overlap))
    overlap_upper_frame = numpy.take_along_axis(
        upper_proj,
        o_upper_sel[:, None] if axis_proj_space == 0 else o_upper_sel[None, :],
        axis=axis_proj_space,
    )
    o_lower_sel = numpy.array(range(start_overlap, end_overlap))
    overlap_lower_frame = numpy.take_along_axis(
        lower_proj,
        o_lower_sel[:, None] if axis_proj_space == 0 else o_upper_sel[None, :],
        axis=axis_proj_space,
    )

    if not overlap_upper_frame.shape == overlap_lower_frame.shape:
        raise ValueError(f"Fail to get consistent overlap ({overlap_upper_frame.shape} vs {overlap_lower_frame.shape})")

    return find_frame_relative_shifts(
        overlap_upper_frame=overlap_upper_frame,
        overlap_lower_frame=overlap_lower_frame,
        estimated_shifts=estimated_shifts,
        x_cross_correlation_function=x_cross_correlation_function,
        y_cross_correlation_function=y_cross_correlation_function,
        x_shifts_params=x_shifts_params,
        y_shifts_params=y_shifts_params,
        overlap_axis=axis_proj_space,
    )


def find_shift_correlate(img1, img2, padding_mode="reflect"):
    alignment = AlignmentBase()
    cc = alignment._compute_correlation_fft(
        img1,
        img2,
        padding_mode,
    )

    img_shape = img1.shape[-2:]
    cc_vs = numpy.fft.fftfreq(img_shape[-2], 1 / img_shape[-2])
    cc_hs = numpy.fft.fftfreq(img_shape[-1], 1 / img_shape[-1])

    (f_vals, fv, fh) = alignment.extract_peak_region_2d(cc, cc_vs=cc_vs, cc_hs=cc_hs)
    shifts_vh = alignment.refine_max_position_2d(f_vals, fv, fh)
    return -shifts_vh


def find_shift_with_itk(img1: numpy.ndarray, img2: numpy.ndarray) -> tuple:
    # created from https://examples.itk.org/src/registration/common/perform2dtranslationregistrationwithmeansquares/documentation
    # return (y_shift, x_shift). For now shift are integers as only integer shift are handled.
    if not img1.dtype == img2.dtype:
        raise ValueError("the two images are expected to have the same type")
    if not img1.ndim == img2.ndim == 2:
        raise ValueError("the two images are expected to 2D numpy arrays")

    if not has_itk:
        _logger.warning("itk is not installed. Please install it to find shift with it")
        return (0, 0)

    if StrictVersion(itk.Version.GetITKVersion()) < StrictVersion("4.9.0"):
        _logger.error("ITK 4.9.0 is required to find shift with it.")
        return (0, 0)

    pixel_type = itk.ctype("float")
    img1 = numpy.ascontiguousarray(img1, dtype=numpy.float32)
    img2 = numpy.ascontiguousarray(img2, dtype=numpy.float32)

    dimension = 2
    image_type = itk.Image[pixel_type, dimension]

    fixed_image = itk.PyBuffer[image_type].GetImageFromArray(img1)
    moving_image = itk.PyBuffer[image_type].GetImageFromArray(img2)

    transform_type = itk.TranslationTransform[itk.D, dimension]
    initial_transform = transform_type.New()

    optimizer = itk.RegularStepGradientDescentOptimizerv4.New(
        LearningRate=4,
        MinimumStepLength=0.001,
        RelaxationFactor=0.5,
        NumberOfIterations=200,
    )

    metric = itk.MeanSquaresImageToImageMetricv4[image_type, image_type].New()

    registration = itk.ImageRegistrationMethodv4.New(
        FixedImage=fixed_image,
        MovingImage=moving_image,
        Metric=metric,
        Optimizer=optimizer,
        InitialTransform=initial_transform,
    )

    moving_initial_transform = transform_type.New()
    initial_parameters = moving_initial_transform.GetParameters()
    initial_parameters[0] = 0
    initial_parameters[1] = 0
    moving_initial_transform.SetParameters(initial_parameters)
    registration.SetMovingInitialTransform(moving_initial_transform)

    identity_transform = transform_type.New()
    identity_transform.SetIdentity()
    registration.SetFixedInitialTransform(identity_transform)

    registration.SetNumberOfLevels(1)
    registration.SetSmoothingSigmasPerLevel([0])
    registration.SetShrinkFactorsPerLevel([1])

    registration.Update()

    transform = registration.GetTransform()
    final_parameters = transform.GetParameters()
    translation_along_x = final_parameters.GetElement(0)
    translation_along_y = final_parameters.GetElement(1)

    return numpy.round(translation_along_y), numpy.round(translation_along_x)


def from_slice_to_n_elements(slice_: Union[slice, tuple]):
    """Return the number of element in a slice or in a tuple"""
    if isinstance(slice_, slice):
        return (slice_.stop - slice_.start) / (slice_.step or 1)
    else:
        return len(slice_)



from ...reconstruction.fbp import *
from .filtering import HelicalSinoFilter
from ...utils import convert_index


class BackprojectorHelical(Backprojector):
    """This is the Backprojector derived class for helical reconstruction.
    The modifications are detailed here :

         * the backprojection is decoupled from the filtering. This allows, in the pipeline,
           for doing first a filtering using backprojector_object.sino_filter subobject,
           then calling backprojector_object.backprojection only after reweigthing the result of
           the filter_sino method of sino_filter subobject.
         * the angles can be resetted on the fly, and the class can work with a variable number of projection.
           As a matter of fact, with the helical_chunked_regridded.py pipeline, the reconstruction is done
           each time with the same set of angles, this because of the regridding mechanism. The feature
           might return useful in the future if alternative approachs are developed again.
         *
    """

    def __init__(self, *args, **kwargs):
        """This became needed after the _d_sino allocation was removed from the base class"""
        super().__init__(*args, **kwargs)
        self._d_sino = self._processing.allocate_array("d_sino", self.sino_shape, "f")

    def set_custom_angles_and_axis_corrections(self, angles_rad, x_per_proj):
        """To arbitrarily change angles
        Parameters
        ==========

        angles_rad: array of floats
           one angle per each projection in radians

        x_per_proj: array of floats
           each entry is the axis shift for a projection, in pixels.

        """
        self.n_provided_angles = len(angles_rad)
        self._axis_correction = np.zeros((1, self.n_angles), dtype=np.float32)

        self._axis_correction[0, : self.n_provided_angles] = -x_per_proj

        self.angles[: self.n_provided_angles] = angles_rad

        self._compute_angles_again()

        self.kern_proj_args[1] = self.n_provided_angles
        self.kern_proj_args[6] = self.offsets["x"]
        self.kern_proj_args[7] = self.offsets["y"]

    def _compute_angles_again(self):
        """to update angles dependent auxiliary arrays"""
        self.h_cos[0] = np.cos(self.angles).astype("f")
        self.h_msin[0] = (-np.sin(self.angles)).astype("f")
        self._d_msin[:] = self.h_msin[0]
        self._d_cos[:] = self.h_cos[0]
        if self._axis_correction is not None:
            self._d_axcorr[:] = self._axis_correction

    def _init_filter(self, filter_name):
        """To use the HelicalSinoFilter which is derived from SinoFilter
        with a slightly different padding scheme
        """
        self.filter_name = filter_name
        self.sino_filter = HelicalSinoFilter(
            self.sino_shape,
            filter_name=self.filter_name,
            padding_mode=self.padding_mode,
            cuda_options={"ctx": self.cuda_processing.ctx},
        )

    def backprojection(self, sino, output=None):
        """Redefined here to do backprojection only, compare to the base class method."""
        self._d_sino[:] = sino
        res = self.backproj(self._d_sino, output=output)
        return res

    def _init_geometry(self, sino_shape, slice_shape, angles, rot_center, slice_roi):
        """this is identical to _init_geometry of the base class with the
        exception that self.extra_options["centered_axis"] is not considered
        and as a consequence  self.offsets is not set here and the one of  _set_slice_roi
        is kept.
        """
        if slice_shape is not None and slice_roi is not None:
            raise ValueError("slice_shape and slice_roi cannot be used together")
        self.sino_shape = sino_shape
        if len(sino_shape) == 2:
            n_angles, dwidth = sino_shape
        else:
            raise ValueError("Expected 2D sinogram")
        self.dwidth = dwidth
        self.rot_center = rot_center or (self.dwidth - 1) / 2.0
        self._set_slice_shape(
            slice_shape,
        )
        self.axis_pos = self.rot_center
        self._set_angles(angles, n_angles)
        self._set_slice_roi(slice_roi)
        self._set_axis_corr()

    def _set_slice_shape(self, slice_shape):
        """this is identical to the _set_slice_shape ofthe base class
        with the exception that n_y,n_x default to the largest possible reconstructible slice
        """

        n_y = self.dwidth + abs(self.dwidth - 1 - self.rot_center * 2)
        n_x = self.dwidth + abs(self.dwidth - 1 - self.rot_center * 2)

        if slice_shape is not None:
            if np.isscalar(slice_shape):
                slice_shape = (slice_shape, slice_shape)
            n_y, n_x = slice_shape

        self.n_x = n_x
        self.n_y = n_y
        self.slice_shape = (n_y, n_x)

    def _set_slice_roi(self, slice_roi):
        """Automatically tune the offset to in all cases."""
        self.slice_roi = slice_roi
        if slice_roi is None:
            off = -(self.dwidth - 1 - self.rot_center * 2)
            if off < 0:
                self.offsets = {"x": off, "y": off}
            else:
                self.offsets = {"x": 0, "y": 0}
        else:
            start_x, end_x, start_y, end_y = slice_roi
            # convert negative indices
            slice_width, _ = self.slice_shape

            off = min(0, -(self.dwidth - 1 - self.rot_center * 2))

            if end_x < start_x:
                start_x = off
                end_x = off + slice_width

            if end_y < start_y:
                start_y = off
                end_y = off + slice_width

            self.slice_shape = (end_y - start_y, end_x - start_x)

            self.n_x = self.slice_shape[-1]
            self.n_y = self.slice_shape[-2]

            self.offsets = {"x": start_x, "y": start_y}



from os import path, environ
from glob import glob
from ..utils import get_folder_path
from ..pipeline.config import generate_nabu_configfile, parse_nabu_config_file
from ..pipeline.fullfield.nabu_config import nabu_config as default_fullfield_config
from ..pipeline.helical.nabu_config import nabu_config as helical_fullfield_config
from .utils import parse_params_values
from .cli_configs import BootstrapConfig


def bootstrap():
    args = parse_params_values(BootstrapConfig, parser_description="Initialize a nabu configuration file")

    do_bootstrap = bool(args["bootstrap"])
    no_comments = bool(args["nocomments"])
    overwrite = bool(args["overwrite"])

    if do_bootstrap:
        print(
            "The --bootstrap option is now the default behavior of the nabu-config command. This option is therefore not needed anymore."
        )
    if path.isfile(args["output"]) and not (overwrite):
        rep = input("File %s already exists. Overwrite ? [y/N]" % args["output"])
        if rep.lower() != "y":
            print("Stopping")
            exit(0)

    opts_level = args["level"]

    prefilled_values = {}
    template_name = args["template"]
    if template_name != "":
        prefilled_values = get_config_template(template_name, if_not_found="print")
        if prefilled_values is None:
            exit(0)
        opts_level = "advanced"

    if args["dataset"] != "":
        prefilled_values["dataset"] = {}
        user_dataset = args["dataset"]
        if not path.isabs(user_dataset):
            user_dataset = path.abspath(user_dataset)
            print("Warning: using absolute dataset path %s" % user_dataset)
        if not path.exists(user_dataset):
            print("Error: cannot find the file or directory %s" % user_dataset)
            exit(1)
        prefilled_values["dataset"]["location"] = user_dataset

    if args["helical"]:
        my_config = helical_fullfield_config
    else:
        my_config = default_fullfield_config

    generate_nabu_configfile(
        args["output"],
        my_config,
        comments=not (no_comments),
        options_level=opts_level,
        prefilled_values=prefilled_values,
    )
    return 0


def get_config_template(template_name, if_not_found="raise"):
    def handle_not_found(msg):
        if if_not_found == "raise":
            raise FileNotFoundError(msg)
        elif if_not_found == "print":
            print(msg)

    templates_path = get_folder_path(path.join("resources", "templates"))

    custom_templates_path = environ.get("NABU_TEMPLATES_PATH", None)
    templates = glob(path.join(templates_path, "*.conf"))
    if custom_templates_path is not None:
        templates_custom = glob(path.join(custom_templates_path, "*.conf"))
        templates_custom += glob(path.join(custom_templates_path, "*.cfg"))
        templates = templates_custom + templates

    available_templates_names = [path.splitext(path.basename(fname))[0] for fname in templates]

    if template_name not in available_templates_names:
        handle_not_found("Unable to find template '%s'. Available are: %s" % (template_name, available_templates_names))
        return

    fname = templates[available_templates_names.index(template_name)]
    return parse_nabu_config_file(fname)



from ..processing.fftshift import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.fftshift has been moved to nabu.processing.fftshift", do_print=True, func_name="fftshift"
)



from multiprocessing.pool import ThreadPool
import numpy as np

from nabu.utils import get_num_threads
from ..misc.binning import binning as image_binning
from ..io.reader import NXTomoReader, EDFStackReader


#
# NXTomoReader with binning
#
def bin_image_stack(src_stack, dst_stack, binning_factor=(2, 2), num_threads=8):
    def _apply_binning(img_res_tuple):
        img, res = img_res_tuple
        res[:] = image_binning(img, binning_factor)

    if dst_stack is None:
        dst_stack = np.zeros((src_stack.shape[0],) + image_binning(src_stack[0], binning_factor).shape, dtype="f")

    with ThreadPool(num_threads) as tp:
        tp.map(_apply_binning, zip(src_stack, dst_stack))
    return dst_stack


def NXTomoReaderBinning(binning_factor, *nxtomoreader_args, num_threads=None, **nxtomoreader_kwargs):
    [
        nxtomoreader_kwargs.pop(kwarg, None)
        for kwarg in ["processing_func", "processing_func_args", "processing_func_kwargs"]
    ]
    nxtomoreader_kwargs["processing_func"] = bin_image_stack
    nxtomoreader_kwargs["processing_func_kwargs"] = {
        "binning_factor": binning_factor,
        "num_threads": num_threads or get_num_threads(),
    }

    return NXTomoReader(
        *nxtomoreader_args,
        **nxtomoreader_kwargs,
    )


#
# NXTomoReader with distortion correction
#


def apply_distortion_correction_on_images_stack(src_stack, dst_stack, distortion_corrector, num_threads=8):
    _, subregion = distortion_corrector.get_actual_shapes_source_target()
    src_x_start, src_x_end, src_z_start, src_z_end = subregion
    if dst_stack is None:
        dst_stack = np.zeros([src_stack.shape[0], src_z_end - src_z_start, src_x_end - src_x_start], "f")

    def apply_corrector(i_img_tuple):
        i, img = i_img_tuple
        dst_stack[i] = distortion_corrector.transform(img)

    with ThreadPool(num_threads) as tp:
        tp.map(apply_corrector, enumerate(src_stack))

    return dst_stack


def NXTomoReaderDistortionCorrection(distortion_corrector, *nxtomoreader_args, num_threads=None, **nxtomoreader_kwargs):
    [
        nxtomoreader_kwargs.pop(kwarg, None)
        for kwarg in ["processing_func", "processing_func_args", "processing_func_kwargs"]
    ]
    nxtomoreader_kwargs["processing_func"] = apply_distortion_correction_on_images_stack
    nxtomoreader_kwargs["processing_func_args"] = [distortion_corrector]
    nxtomoreader_kwargs["processing_func_kwargs"] = {"num_threads": num_threads or get_num_threads()}

    return NXTomoReader(
        *nxtomoreader_args,
        **nxtomoreader_kwargs,
    )


#
# EDF Reader with binning
#


def EDFStackReaderBinning(binning_factor, *edfstackreader_args, **edfstackreader_kwargs):
    [
        edfstackreader_kwargs.pop(kwarg, None)
        for kwarg in ["processing_func", "processing_func_args", "processing_func_kwargs"]
    ]
    edfstackreader_kwargs["processing_func"] = image_binning
    edfstackreader_kwargs["processing_func_args"] = [binning_factor]

    return EDFStackReader(
        *edfstackreader_args,
        **edfstackreader_kwargs,
    )


#
# EDF Reader with distortion correction
#


def apply_distortion_correction_on_image(image, distortion_corrector):
    return distortion_corrector.transform(image)


def EDFStackReaderDistortionCorrection(distortion_corrector, *edfstackreader_args, **edfstackreader_kwargs):
    [
        edfstackreader_kwargs.pop(kwarg, None)
        for kwarg in ["processing_func", "processing_func_args", "processing_func_kwargs"]
    ]
    edfstackreader_kwargs["processing_func"] = apply_distortion_correction_on_image
    edfstackreader_kwargs["processing_func_args"] = [distortion_corrector]

    return EDFStackReader(
        *edfstackreader_args,
        **edfstackreader_kwargs,
    )


def load_darks_flats(
    dataset_info, sub_region, processing_func=None, processing_func_args=None, processing_func_kwargs=None
):
    """
    Load the (reduced) darks and flats and crop them to the sub-region currently used.
    At this stage, dataset_info.flats should be a dict in the form {num: array}

    Parameters
    ----------
    sub_region: 2-tuple of 3-tuples of int
        Tuple in the form ((start_y, end_y), (start_x, end_x))
    """
    (start_y, end_y), (start_x, end_x) = sub_region

    processing_func_args = processing_func_args or []
    processing_func_kwargs = processing_func_kwargs or {}

    def proc(img):
        if processing_func is None:
            return img
        return processing_func(img, *processing_func_args, **processing_func_kwargs)

    res = {
        "flats": {k: proc(flat_k)[start_y:end_y, start_x:end_x] for k, flat_k in dataset_info.flats.items()},
        "darks": {k: proc(dark_k)[start_y:end_y, start_x:end_x] for k, dark_k in dataset_info.darks.items()},
    }
    return res



from ..processing.medfilt_cuda import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.cuda.medfilt has been moved to nabu.processing.medfilt_cuda", do_print=True, func_name="medfilt_cuda"
)



from multiprocessing.pool import ThreadPool
import numpy as np

try:
    from skimage.transform import warp_polar

    __have_skimage__ = True
except ImportError:
    __have_skimage__ = False


def azimuthal_integration(img, axes=(-2, -1), domain="direct"):
    """
    Computes azimuthal integration of an image or a stack of images.

    Parameters
    ----------
    img : `numpy.array_like`
        The image or stack of images.
    axes : tuple(int, int), optional
        Axes of that need to be azimuthally integrated. The default is (-2, -1).
    domain : string, optional
        Domain of the integration. Options are: "direct" | "fourier". Default is "direct".

    Raises
    ------
    ValueError
        Error returned when not passing images or wrong axes.
    NotImplementedError
        In case of tack of images for the moment.

    Returns
    -------
    `numpy.array_like`
        The azimuthally integrated profile.
    """
    if not len(img.shape) >= 2:
        raise ValueError("Input image should be at least 2-dimensional.")
    if not len(axes) == 2:
        raise ValueError("Input axes should be 2.")

    img_axes_dims = np.array((img.shape[axes[0]], img.shape[axes[1]]))
    if domain.lower() == "direct":
        half_dims = (img_axes_dims - 1) / 2
        xx = np.linspace(-half_dims[0], half_dims[0], img_axes_dims[0])
        yy = np.linspace(-half_dims[1], half_dims[1], img_axes_dims[1])
    else:
        xx = np.fft.fftfreq(img_axes_dims[0], 1 / img_axes_dims[0])
        yy = np.fft.fftfreq(img_axes_dims[1], 1 / img_axes_dims[1])
    xy = np.stack(np.meshgrid(xx, yy, indexing="ij"))
    r = np.sqrt(np.sum(xy**2, axis=0))

    img_tr_op = [*range(len(img.shape))]
    for a in axes:
        img_tr_op.append(img_tr_op.pop(a))
    img = np.transpose(img, img_tr_op)
    if len(img.shape) > 2:
        img_old_shape = img.shape[:-2]
        img = np.reshape(img, [-1, *img_axes_dims])

    r_l = np.floor(r)
    r_u = r_l + 1
    w_l = (r_u - r) * img
    w_u = (r - r_l) * img

    r_all = np.concatenate((r_l.flatten(), r_u.flatten())).astype(np.int64)
    if len(img.shape) == 2:
        w_all = np.concatenate((w_l.flatten(), w_u.flatten()))
        return np.bincount(r_all, weights=w_all)
    else:
        num_imgs = img.shape[0]
        az_img = [None] * num_imgs
        for ii in range(num_imgs):
            w_all = np.concatenate((w_l[ii, :].flatten(), w_u[ii, :].flatten()))
            az_img[ii] = np.bincount(r_all, weights=w_all)
        az_img = np.array(az_img)
        return np.reshape(az_img, (*img_old_shape, az_img.shape[-1]))


def do_radial_distribution(ip, X0, Y0, mR, nBins=None, use_calibration=False, cal=None, return_radii=False):
    """
    Translates the Java method `doRadialDistribution` (from imagej) into Python using NumPy.
    Done by chatgpt-4o on 2024-11-08

    Args:
    - ip: A 2D numpy array representing the image.
    - X0, Y0: Coordinates of the center.
    - mR: Maximum radius.
    - nBins: Number of bins (optional, defaults to 3*mR/4).
    - use_calibration: Boolean indicating if calibration should be applied.
    - cal: Calibration object with attributes `pixel_width` and `units` (optional).
    """
    if nBins is None:
        nBins = int(3 * mR / 4)

    Accumulator = np.zeros((2, nBins))

    # Define the bounding box
    xmin, xmax = X0 - mR, X0 + mR
    ymin, ymax = Y0 - mR, Y0 + mR

    # Create grid of coordinates
    x = np.arange(xmin, xmax)
    y = np.arange(ymin, ymax)
    xv, yv = np.meshgrid(x, y, indexing="ij")

    # Calculate the radius for each point
    R = np.sqrt((xv - X0) ** 2 + (yv - Y0) ** 2)

    # Bin calculation
    bins = np.floor((R / mR) * nBins).astype(int)
    bins = np.clip(bins - 1, 0, nBins - 1)  # Adjust bins to be in range [0, nBins-1]

    # Accumulate values
    for b in range(nBins):
        mask = bins == b
        Accumulator[0, b] = np.sum(mask)
        Accumulator[1, b] = np.sum(ip[mask])

    # Normalize integrated intensity
    Accumulator[1] /= Accumulator[0]

    if use_calibration and cal is not None:
        # Apply calibration if units are provided
        radii = cal.pixel_width * mR * (np.arange(1, nBins + 1) / nBins)
        units = cal.units
    else:
        # Use pixel units
        radii = mR * (np.arange(1, nBins + 1) / nBins)
        units = "pixels"

    if return_radii:
        return radii, Accumulator[1]
    else:
        return Accumulator[1]


# OK-ish, but small discrepancy with do_radial_distribution.
# 20-40X faster than above methods for (2048, 2048) images
# Also it assumes a uniform sampling
# No idea why there is this "offset=1", to be investigated - perhaps radius=0 is also calculated ?
def azimuthal_integration_skimage(img, center=None, offset=1):
    shape2 = [int(s // 2 * 1.4142) for s in img.shape]
    s = min(img.shape) // 2
    img_polar = warp_polar(img, output_shape=shape2, center=center)
    return img_polar.mean(axis=0)[offset : offset + s]


def _apply_on_images_stack(func, images_stack, n_threads=4, func_args=None, func_kwargs=None):
    func_args = func_args or []
    func_kwargs = func_kwargs or {}

    def _process_image(img):
        return func(img, *func_args, **func_kwargs)

    with ThreadPool(n_threads) as tp:
        res = tp.map(_process_image, images_stack)
    return np.array(res)


def _apply_on_patches_stack(func, images_stack, n_threads=4, func_args=None, func_kwargs=None):
    (n_images, n_patchs_y, img_shape_y, n_patchs_x, img_shape_x) = images_stack.shape
    func_args = func_args or []
    func_kwargs = func_kwargs or {}
    out_sample = func(images_stack[0, 0, :, 0, :], *func_args, **func_kwargs)
    out_shape = out_sample.shape
    out_dtype = out_sample.dtype

    def _process_image(img):
        res = np.zeros((n_patchs_y, n_patchs_x) + out_shape, dtype=out_dtype)
        for i in range(n_patchs_y):
            for j in range(n_patchs_x):
                res[i, j] = func(img[i, :, j, :], *func_args, **func_kwargs)
        return res

    with ThreadPool(n_threads) as tp:
        res = tp.map(_process_image, images_stack)
    return np.array(res)


def azimuthal_integration_imagej_stack(images_stack, n_threads=4):
    if images_stack.ndim == 3:
        img_shape = images_stack.shape[-2:]
        _apply = _apply_on_images_stack
    elif images_stack.ndim == 5:
        img_shape = np.array(images_stack.shape)[[-3, -1]]
        _apply = _apply_on_patches_stack
    else:
        raise ValueError
    s = min(img_shape)
    return _apply(
        do_radial_distribution,
        images_stack,
        n_threads=n_threads,
        func_args=[s // 2, s // 2, s // 2],
        func_kwargs={"nBins": s // 2, "return_radii": False},
    )


def azimuthal_integration_skimage_stack(images_stack, n_threads=4):
    if images_stack.ndim == 3:
        return _apply_on_images_stack(azimuthal_integration_skimage, images_stack, n_threads=n_threads)
    elif images_stack.ndim == 5:
        return _apply_on_patches_stack(azimuthal_integration_skimage, images_stack, n_threads=n_threads)
    else:
        raise ValueError



import pytest
import numpy as np
from nabu.preproc.phase import PaganinPhaseRetrieval
from nabu.processing.fft_cuda import get_available_fft_implems
from nabu.testutils import generate_tests_scenarios, get_data
from nabu.thirdparty.tomopy_phase import retrieve_phase
from nabu.cuda.utils import __has_pycuda__

__has_cufft__ = False
if __has_pycuda__:
    from nabu.preproc.phase_cuda import CudaPaganinPhaseRetrieval

    avail_fft = get_available_fft_implems()
    __has_cufft__ = len(avail_fft) > 0

scenarios = {
    "distance": [1],
    "energy": [35],
    "delta_beta": [1e1],
    "margin": [((50, 50), (0, 0)), None],
}

scenarios = generate_tests_scenarios(scenarios)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    cls.data = get_data("mri_proj_astra.npz")["data"]
    cls.rtol = 1.1e-6
    cls.rtol_pag = 5e-3


@pytest.mark.usefixtures("bootstrap")
class TestPaganin:
    """
    Test the Paganin phase retrieval.
    The reference implementation is tomopy.
    """

    @staticmethod
    def get_paganin_instance_and_data(cfg, data):
        pag_kwargs = cfg.copy()
        margin = pag_kwargs.pop("margin")
        if margin is not None:
            data = np.pad(data, margin, mode="edge")
        paganin = PaganinPhaseRetrieval(data.shape, **pag_kwargs)
        return paganin, data, pag_kwargs

    @staticmethod
    def crop_to_margin(data, margin):
        if margin is None:
            return data
        ((U, D), (L, R)) = margin
        D = None if D == 0 else -D
        R = None if R == 0 else -R
        return data[U:D, L:R]

    @pytest.mark.parametrize("config", scenarios)
    def test_paganin(self, config):
        paganin, data, _ = self.get_paganin_instance_and_data(config, self.data)
        res = paganin.apply_filter(data)

        data_tomopy = np.atleast_3d(np.copy(data)).T
        res_tomopy = retrieve_phase(
            data_tomopy,
            pixel_size=paganin.pixel_size_xy_micron[0] * 1e-4,
            dist=paganin.distance_cm,
            energy=paganin.energy_kev,
            alpha=1.0 / (4 * 3.141592**2 * paganin.delta_beta),
        )

        res_tomopy = self.crop_to_margin(res_tomopy[0].T, config["margin"])
        res = self.crop_to_margin(res, config["margin"])

        errmax = np.max(np.abs(res - res_tomopy) / np.max(res_tomopy))
        assert errmax < self.rtol_pag, "Max error is too high"

    @pytest.mark.skipif(
        not (__has_pycuda__ and __has_cufft__), reason="Need pycuda and (scikit-cuda or vkfft) for this test"
    )
    @pytest.mark.parametrize("config", scenarios)
    def test_gpu_paganin(self, config):
        paganin, data, pag_kwargs = self.get_paganin_instance_and_data(config, self.data)

        gpu_paganin = CudaPaganinPhaseRetrieval(data.shape, **pag_kwargs)
        ref = paganin.apply_filter(data)
        res = gpu_paganin.apply_filter(data)
        errmax = np.max(np.abs((res - ref) / np.max(ref)))
        assert errmax < self.rtol, "Max error is too high"



from ..utils import MissingComponentError
from ..processing.processing_base import ProcessingBase
from .utils import get_cuda_context, __has_pycuda__

if __has_pycuda__:
    import pycuda.driver as cuda
    import pycuda.gpuarray as garray
    from ..cuda.kernel import CudaKernel

    dev_attrs = cuda.device_attribute
    GPUArray = garray.GPUArray
    from pycuda.tools import dtype_to_ctype
else:
    GPUArray = MissingComponentError("pycuda")
    dtype_to_ctype = MissingComponentError("pycuda")


# NB: we must detach from a context before creating another context
class CudaProcessing(ProcessingBase):
    array_class = GPUArray if __has_pycuda__ else None
    dtype_to_ctype = dtype_to_ctype

    def __init__(self, device_id=None, ctx=None, stream=None, cleanup_at_exit=True):
        """
        Initialie a CudaProcessing instance.

        CudaProcessing is a base class for all CUDA-based processings.
        This class provides utilities for context/device management, and
        arrays allocation.

        Parameters
        ----------
        device_id: int, optional
            ID of the cuda device to use (those of the `nvidia-smi` command).
            Ignored if ctx is not None.
        ctx: pycuda.driver.Context, optional
            Existing context to use. If provided, do not create a new context.
        stream: pycudacuda.driver.Stream, optional
            Cuda stream. If not provided, will use the default stream
        cleanup_at_exit: bool, optional
            Whether to clean-up the context at exit.
            Ignored if ctx is not None.
        """
        super().__init__()
        if ctx is None:
            self.ctx = get_cuda_context(device_id=device_id, cleanup_at_exit=cleanup_at_exit)
        else:
            self.ctx = ctx
        self.stream = stream
        self.device = self.ctx.get_device()
        self.device_name = self.device.name()
        self.device_id = self.device.get_attribute(dev_attrs.MULTI_GPU_BOARD_GROUP_ID)  # pylint: disable=E0606

    def push_context(self):
        self.ctx.push()
        return self.ctx

    def pop_context(self):
        self.ctx.pop()

    def _allocate_array_mem(self, shape, dtype):
        return garray.zeros(shape, dtype)

    def kernel(
        self, kernel_name, filename=None, src=None, signature=None, texrefs=None, automation_params=None, **build_kwargs
    ):
        return CudaKernel(  # pylint: disable=E0606
            kernel_name,
            filename=filename,
            src=src,
            signature=signature,
            texrefs=texrefs,
            automation_params=automation_params,
            **build_kwargs,
        )



import numpy as np
import pycuda.driver as cuda
import pycuda.gpuarray as garray
from ..utils import updiv, get_cuda_srcfile
from ..cuda.utils import copy_array
from ..cuda.kernel import CudaKernel
from ..cuda.processing import CudaProcessing

_sizeof_float32 = np.dtype(np.float32).itemsize


class Projector:
    """
    A class for performing a tomographic projection (Radon Transform) using Cuda.
    """

    _projector_name = "joseph_projector"
    _projector_signature = "PiiPfPPPPiiifii"

    def __init__(
        self,
        slice_shape,
        angles,
        rot_center=None,
        detector_width=None,
        normalize=False,
        extra_options=None,
        cuda_options=None,
    ):
        """
        Initialize a Cuda tomography forward projector.

        Parameters
        -----------

        slice_shape: tuple
            Shape of the slice: (num_rows, num_columns).
        angles: int or sequence
            Either an integer number of angles, or a list of custom angles values in radian.
        param rot_center: float, optional
            Rotation axis position. Default is `(shape[1]-1)/2.0`.
        detector_width: int, optional
            Detector width in pixels.
            If `detector_width > slice_shape[1]`, the  projection data will be surrounded with zeros.
            Using `detector_width < slice_shape[1]` might result in a local tomography setup.
        normalize: bool, optional
            Whether to normalize projection.
            If set to True, sinograms are multiplied by the factor pi/(2*nprojs).
        extra_options: dict, optional
            Current allowed options:
                offset_x, axis_corrections
        cuda_options: dict, optional
            Cuda options passed to the CudaProcessing class.
        """
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._configure_extra_options(extra_options)
        self._init_geometry(slice_shape, rot_center, angles, detector_width)
        self.normalize = normalize
        self._allocate_memory()
        self._compute_angles()
        self._proj_precomputations()
        self._compile_kernels()

    def _configure_extra_options(self, extra_options):
        self.extra_options = {
            "offset_x": None,
            "axis_corrections": None,  # TODO
        }
        extra_opts = extra_options or {}
        self.extra_options.update(extra_opts)

    def _init_geometry(self, slice_shape, rot_center, angles, detector_width):
        if np.isscalar(slice_shape):
            slice_shape = (slice_shape, slice_shape)
        self.shape = slice_shape
        if np.isscalar(angles):
            angles = np.linspace(0, np.pi, angles, endpoint=False, dtype="f")
        self.angles = angles
        self.nprojs = len(angles)
        self.dwidth = detector_width or self.shape[1]
        self.sino_shape = (self.nprojs, self.dwidth)

        # In PYHST (c_hst_project_1over.cu), axis_pos is overwritten to (dimslice-1)/2.
        # So tuning axis position is done in another way. In CCspace.c:
        # offset_x = start_x - move_x
        # start_x = start_voxel_1 (zero-based, so 0 by default)
        # MOVE_X = start_x + (num_x - 1)/2 -  ROTATION_AXIS_POSITION;
        self.axis_pos = self.rot_center = rot_center or (self.shape[1] - 1) / 2.0
        self.offset_x = self.extra_options["offset_x"] or np.float32(self.axis_pos - (self.shape[1] - 1) / 2.0)
        self.axis_pos0 = np.float32((self.shape[1] - 1) / 2.0)

    def _allocate_memory(self):
        self.dimgrid_x = updiv(self.dwidth, 16)
        self.dimgrid_y = updiv(self.nprojs, 16)
        self._dimrecx = self.dimgrid_x * 16
        self._dimrecy = self.dimgrid_y * 16
        self.d_sino = garray.zeros((self._dimrecy, self._dimrecx), np.float32)
        self.d_angles = garray.zeros((self._dimrecy,), np.float32)
        self._d_beginPos = garray.zeros((2, self._dimrecy), np.int32)
        self._d_strideJoseph = garray.zeros((2, self._dimrecy), np.int32)
        self._d_strideLine = garray.zeros((2, self._dimrecy), np.int32)
        self.d_axis_corrections = garray.zeros((self.nprojs,), np.float32)
        if self.extra_options.get("axis_corrections", None) is not None:
            self.d_axis_corrections.set(self.extra_options["axis_corrections"])

        # Textures
        self.d_image_cua = cuda.np_to_array(np.zeros((self.shape[0] + 2, self.shape[1] + 2), "f"), "C")

    def _compile_kernels(self):
        self.gpu_projector = CudaKernel(
            self._projector_name,
            filename=get_cuda_srcfile("proj.cu"),
        )
        self.texref_slice = self.gpu_projector.module.get_texref("texSlice")
        self.texref_slice.set_array(self.d_image_cua)
        self.texref_slice.set_filter_mode(cuda.filter_mode.LINEAR)
        self.gpu_projector.prepare(self._projector_signature, [self.texref_slice])

        self.kernel_args = (
            self.d_sino.gpudata,
            np.int32(self.shape[1]),
            np.int32(self.dwidth),
            self.d_angles.gpudata,
            np.float32(self.axis_pos0),
            self.d_axis_corrections.gpudata,
            self._d_beginPos.gpudata,
            self._d_strideJoseph.gpudata,
            self._d_strideLine.gpudata,
            np.int32(self.nprojs),
            np.int32(self._dimrecx),
            np.int32(self._dimrecy),
            self.offset_x,
            np.int32(1),  # josephnoclip, 1 by default
            np.int32(self.normalize),
        )
        self._proj_kernel_blk = (16, 16, 1)
        self._proj_kernel_grd = (self.dimgrid_x, self.dimgrid_y, 1)

    def _compute_angles(self):
        angles2 = np.zeros(self._dimrecy, dtype=np.float32)  # dimrecy != num_projs
        angles2[: self.nprojs] = np.copy(self.angles)
        angles2[self.nprojs :] = angles2[self.nprojs - 1]
        self.angles2 = angles2
        self.d_angles[:] = angles2[:]

    def _proj_precomputations(self):
        beginPos = np.zeros((2, self._dimrecy), dtype=np.int32)
        strideJoseph = np.zeros((2, self._dimrecy), dtype=np.int32)
        strideLine = np.zeros((2, self._dimrecy), dtype=np.int32)
        cos_angles = np.cos(self.angles2)
        sin_angles = np.sin(self.angles2)
        dimslice = self.shape[1]

        M1 = np.abs(cos_angles) > 0.70710678
        M1b = np.logical_not(M1)
        M2 = cos_angles > 0
        M2b = np.logical_not(M2)
        M3 = sin_angles > 0
        M3b = np.logical_not(M3)
        case1 = M1 * M2
        case2 = M1 * M2b
        case3 = M1b * M3
        case4 = M1b * M3b

        beginPos[:, case1] = 0
        strideJoseph[0][case1] = 1
        strideJoseph[1][case1] = 0
        strideLine[0][case1] = 0
        strideLine[1][case1] = 1

        beginPos[:, case2] = dimslice - 1
        strideJoseph[0][case2] = -1
        strideJoseph[1][case2] = 0
        strideLine[0][case2] = 0
        strideLine[1][case2] = -1

        beginPos[0][case3] = dimslice - 1
        beginPos[1][case3] = 0
        strideJoseph[0][case3] = 0
        strideJoseph[1][case3] = 1
        strideLine[0][case3] = -1
        strideLine[1][case3] = 0

        beginPos[0][case4] = 0
        beginPos[1][case4] = dimslice - 1
        strideJoseph[0][case4] = 0
        strideJoseph[1][case4] = -1
        strideLine[0][case4] = 1
        strideLine[1][case4] = 0

        self._d_beginPos.set(beginPos)
        self._d_strideJoseph.set(strideJoseph)
        self._d_strideLine.set(strideLine)

    def _check_input_array(self, image):
        if image.shape != self.shape:
            raise ValueError("Expected slice shape = %s, got %s" % (str(self.shape), str(image.shape)))
        if image.dtype != np.dtype("f"):
            raise ValueError("Expected float32 data type, got %s" % str(image.dtype))
        if not isinstance(image, (np.ndarray, garray.GPUArray)):
            raise ValueError("Expected either numpy.ndarray or pyopencl.array.Array")
        if isinstance(image, np.ndarray):
            if not image.flags["C_CONTIGUOUS"]:
                raise ValueError("Please use C-contiguous arrays")

    def set_image(self, image, check=True):
        if check:
            self._check_input_array(image)
        copy_array(
            self.d_image_cua,
            image,
            dst_x_in_bytes=_sizeof_float32,
            dst_y=1,
            check=False,  # cannot check when using offsets
        )

    def projection(self, image, output=None, do_checks=True):
        """
        Perform the projection of an image.

        Parameters
        -----------
        image: array
            Image to forward project
        output: array, optional
            Output image
        """

        self.set_image(image, check=do_checks)
        self.gpu_projector(*self.kernel_args, grid=self._proj_kernel_grd, block=self._proj_kernel_blk)

        if output is None:
            res = self.d_sino.get()
            res = res[: self.nprojs, : self.dwidth]  # copy ?
        else:
            output[:, :] = self.d_sino[: self.nprojs, : self.dwidth]
            res = output
        return res

    __call__ = projection



from .. import version
from os import environ

import argparse
import shutil
import os
import sys
import re
import h5py
import numpy as np


from ..resources.logger import LoggerOrPrint
from .utils import parse_params_values
from .cli_configs import CorrectRotConfig
from silx.io.dictdump import h5todict

from nxtomo.application.nxtomo import NXtomo

import h5py
from nabu.utils import DictToObj


def main(user_args=None):
    """Applies the correction found by diag_to_rot to a nexus file"""

    if user_args is None:
        user_args = sys.argv[1:]

    args = DictToObj(
        parse_params_values(
            CorrectRotConfig,
            parser_description=main.__doc__,
            program_version="nabu " + version,
            user_args=user_args,
        )
    )

    # now we read the results of the diag_to_rot utility, they are in the cor_file parameter
    # of the cli
    cor_data = DictToObj(h5todict(args.cor_file, "/"))

    my_cor = cor_data.cor[0]
    # we will take my_cor as cor at the first angular position
    # and then we correct the x_translation at the other angles

    # We now load the nexus that we wish to correct
    nx_tomo = NXtomo().load(args.nexus_source, args.entry_name)

    # The cor_file that we use for correction
    # is providing us with the z_m that gives for each
    # cor position found in the cor array the corresponding value of
    # the translation along z (in meters)
    z_translation = nx_tomo.sample.z_translation.value
    z_translation = z_translation - z_translation[0]

    # now we interpolate to find the correction
    # for each position of the encoders
    cors = np.interp(z_translation, cor_data.z_m, cor_data.cor)

    # this is the correction
    x_correction = nx_tomo.instrument.detector.x_pixel_size.value * (cors - my_cor)  # we are in meters here

    # and we apply it to the nexus that we have loaded
    nx_tomo.sample.x_translation = nx_tomo.sample.x_translation.value + x_correction

    # finally we write it to the corrected nexus file
    nx_tomo.save(file_path=args.nexus_target, data_path=args.entry_name, overwrite=True)

    return 0



import os
import numpy
import pytest
from tomoscan.esrf import NXtomoScan
from tomoscan.esrf.volume import HDF5Volume
from tomoscan.esrf.scan.utils import cwd_context
from nabu.stitching.config import PreProcessedZStitchingConfiguration, SlurmConfig
from nabu.stitching.overlap import OverlapStitchingStrategy
from nabu.stitching.slurm_utils import (
    split_slices,
    get_working_directory,
    split_stitching_configuration_to_slurm_job,
)
from tomoscan.esrf.mock import MockNXtomo

try:
    import sluurp
except ImportError:
    has_sluurp = False
else:
    has_sluurp = True


def test_split_slices():
    """test split_slices function"""
    assert tuple(split_slices(slice(0, 100, 1), n_parts=4)) == (
        slice(0, 25, 1),
        slice(25, 50, 1),
        slice(50, 75, 1),
        slice(75, 100, 1),
    )

    assert tuple(split_slices(slice(0, 50, 2), n_parts=3)) == (
        slice(0, 17, 2),
        slice(17, 34, 2),
        slice(34, 50, 2),
    )

    assert tuple(split_slices(slice(0, 100, 1), n_parts=1)) == (slice(0, 100, 1),)

    assert tuple(split_slices(("first", "middle", "last"), 2)) == (
        ("first", "middle"),
        ("last",),
    )
    assert tuple(
        split_slices(
            (
                10,
                12,
                13,
            ),
            4,
        )
    ) == ((10,), (12,), (13,))

    with pytest.raises(TypeError):
        next(split_slices("dsad", 12))


def test_get_working_directory():
    """test get_working_directory function"""
    assert get_working_directory(NXtomoScan("/this/is/my/hdf5file.hdf5", "entry")) == "/this/is/my"
    assert get_working_directory(HDF5Volume("/this/is/my/volume.hdf5", "entry")) == "/this/is/my"


@pytest.mark.skipif(not has_sluurp, reason="sluurp not installed")
def test_split_stitching_configuration_to_slurm_job(tmp_path):
    """
    test split_stitching_configuration_to_slurm_job behavior

    This test is stitching two existing NXtomo (scan1 and scan2 contained in inputs_dir) and create a final_nx_tomo.nx to the output_dir

    The stitching will be split in two slurm jobs.
    One will create output_dir/final_nx_tomo/final_nx_tomo_part_0.nx and the second output_dir/final_nx_tomo/final_nx_tomo_part_1.nx
    then the concatenation (not tested here) will create a output_dir/final_nx_tomo.nx redirecting to the sub parts

    This test only focus on checking each sub configuration is as expected
    """

    inputs_dir = tmp_path / "inputs"
    inputs_dir.mkdir()
    output_dir = tmp_path / "outputs"
    output_dir.mkdir()

    with cwd_context(inputs_dir):
        # the current working directory context help to check file path are moved to absolute.
        # which is important because those jobs will be launched on slurm
        scan_1 = MockNXtomo(
            os.path.join("scan_1"),
            n_proj=10,
            n_ini_proj=10,
            dim=100,
        ).scan

        scan_2 = MockNXtomo(
            os.path.join("scan_2"),
            n_proj=10,
            n_ini_proj=10,
            dim=100,
        ).scan

        n_jobs = 2

        raw_config = PreProcessedZStitchingConfiguration(
            axis_0_pos_px=None,
            axis_0_pos_mm=None,
            axis_0_params={},
            axis_1_pos_px=None,
            axis_1_pos_mm=None,
            axis_1_params={},
            axis_2_pos_px=None,
            axis_2_pos_mm=None,
            axis_2_params={},
            stitching_strategy=OverlapStitchingStrategy.MEAN,
            overwrite_results=True,
            slurm_config=SlurmConfig(
                partition="par-test",
                mem="45G",
                n_jobs=n_jobs,
                other_options="",
                preprocessing_command="source /my/venv",
                clean_script=True,
            ),
            slices=slice(0, 120, 1),
            input_scans=(scan_1, scan_2),
            output_file_path=os.path.join("../outputs/", "final_nx_tomo.nx"),
            output_data_path="stitched_entry",
            output_nexus_version=None,
            slice_for_cross_correlation="middle",
            pixel_size=None,
        )

        sbatch_script_jobs = []
        stitching_configurations = []
        for job, configuration in split_stitching_configuration_to_slurm_job(raw_config, yield_configuration=True):
            sbatch_script_jobs.append(job)
            stitching_configurations.append(configuration)

        assert len(stitching_configurations) == n_jobs == len(sbatch_script_jobs)
        for i_sub_config, sub_config in enumerate(stitching_configurations):
            assert isinstance(sub_config, type(raw_config))
            assert sub_config.slurm_config is None
            assert sub_config.output_file_path == os.path.join(
                output_dir, "final_nx_tomo", f"final_nx_tomo_part_{i_sub_config}.nx"
            )
        assert raw_config.output_file_path == os.path.join(output_dir, "final_nx_tomo.nx")



import numpy as np
import scipy.interpolate
from .translation import DetectorTranslationAlongBeam
from ..misc.filters import correct_spikes
from ..resources.logger import LoggerOrPrint


def estimate_flat_distortion(
    flat,
    image,
    tile_size=100,
    interpolation_kind="linear",
    padding_mode="edge",
    correction_spike_threshold=None,
    logger=None,
):
    """
    Estimate the wavefront distortion on a flat image, from another image.

    Parameters
    ----------
    flat: np.array
        The flat-field image to be corrected
    image: np.ndarray
        The image to correlate the flat against.
    tile_size: int
        The wavefront corrections are calculated by correlating the
        image to the flat, region by region. The regions are tiles of size tile_size
    interpolation_kind: "linear" or "cubic"
        The interpolation method used for interpolation
    padding_mode: string
        Padding mode. Must be valid for np.pad
        when wavefront correction is applied, the corrections are first found for the tiles,
        which gives the shift at the center of each tiled. Then, to interpolate the corrections,
        at the positions f every pixel, on must add also the border of the extremal tiles.
        This is done by padding with a width of 1, and using the mode given 'padding_mode'.
    correction_spike_threshold: float, optional
        By default it is None and no spike correction is performed on the shifts grid which is found by correlation.
        If set to a float, a spike removal will be applied using such threshold

    Returns
    --------
    coordinates: np.ndarray
        An array having dimensions (flat.shape[0], flat.shape[1], 2)
        where each coordinates[i,j] contains the coordinates of the position
        in the  image "flat" which correlates to the pixel (i,j) in the  image "im2".
    """
    logger = LoggerOrPrint(logger)
    starts_r = np.array(range(0, image.shape[0] - tile_size, tile_size))
    starts_c = np.array(range(0, image.shape[1] - tile_size, tile_size))
    cor1 = np.zeros([len(starts_r), len(starts_c)], np.float32)
    cor2 = np.zeros([len(starts_r), len(starts_c)], np.float32)

    shift_finder = DetectorTranslationAlongBeam()
    for ir, r in enumerate(starts_r):
        for ic, c in enumerate(starts_c):
            try:
                coeff_v, coeff_h, shifts_vh_per_img = shift_finder.find_shift(
                    np.array([image[r : r + tile_size, c : c + tile_size], flat[r : r + tile_size, c : c + tile_size]]),
                    np.array([0, 1]),
                    return_shifts=True,
                    low_pass=(1.0, 0.3),
                    high_pass=(tile_size, tile_size * 0.3),
                )
                cor1[ir, ic], cor2[ir, ic] = shifts_vh_per_img[1]

            except ValueError as e:
                if "positions are outside" in str(e):
                    logger.debug(str(e))
                    cor1[ir, ic], cor2[ir, ic] = (0, 0)
                else:
                    raise

    cor1[np.isnan(cor1)] = 0
    cor2[np.isnan(cor2)] = 0

    if correction_spike_threshold is not None:
        cor1 = correct_spikes(cor1, correction_spike_threshold)
        cor2 = correct_spikes(cor2, correction_spike_threshold)

    # TODO implement the previous spikes correction in CCDCorrection - median_clip
    # spikes_corrector = CCDCorrection(cor1.shape, median_clip_thresh=3, abs_diff=True, preserve_borders=True)
    # cor1 = spikes_corrector.median_clip_correction(cor1)
    # cor2 = spikes_corrector.median_clip_correction(cor2)

    cor1 = np.pad(cor1, ((1, 1), (1, 1)), mode=padding_mode)
    cor2 = np.pad(cor2, ((1, 1), (1, 1)), mode=padding_mode)

    hp = np.concatenate([[0.0], starts_c + tile_size * 0.5, [image.shape[1]]])
    vp = np.concatenate([[0.0], starts_r + tile_size * 0.5, [image.shape[0]]])

    h_ticks = np.arange(image.shape[1]).astype(np.float32)
    v_ticks = np.arange(image.shape[0]).astype(np.float32)

    spline_degree = {"linear": 1, "cubic": 3}[interpolation_kind]

    interpolator = scipy.interpolate.RectBivariateSpline(vp, hp, cor1, kx=spline_degree, ky=spline_degree)
    cor1 = interpolator(h_ticks, v_ticks)

    interpolator = scipy.interpolate.RectBivariateSpline(vp, hp, cor2, kx=spline_degree, ky=spline_degree)
    cor2 = interpolator(h_ticks, v_ticks)

    hh = np.arange(image.shape[1]).astype(np.float32)
    vv = np.arange(image.shape[0]).astype(np.float32)
    unshifted_v, unshifted_h = np.meshgrid(vv, hh, indexing="ij")

    shifted_v = unshifted_v - cor1
    shifted_h = unshifted_h - cor2

    coordinates = np.transpose(np.array([shifted_v, shifted_h]), axes=[1, 2, 0])

    return coordinates



#
# Default configuration for CLI tools
#

# Default configuration for "bootstrap" command
from nabu.stitching.definitions import StitchingType
from nabu.pipeline.config_validators import str2bool

from tomoscan.framereducer.method import ReduceMethod


BootstrapConfig = {
    "bootstrap": {
        "help": "DEPRECATED, this is the default behavior. Bootstrap a configuration file from scratch.",
        "action": "store_const",
        "const": 1,
    },
    "convert": {
        "help": "UNSUPPORTED. This option has no effect and will disappear. Convert a PyHST configuration file to a nabu configuration file.",
        "default": "",
    },
    "output": {
        "help": "Output filename",
        "default": "nabu.conf",
    },
    "nocomments": {
        "help": "Remove the comments in the configuration file (default: False)",
        "action": "store_const",
        "const": 1,
    },
    "level": {
        "help": "Level of options to embed in the configuration file. Can be 'required', 'optional', 'advanced'.",
        "default": "optional",
    },
    "dataset": {
        "help": "Pre-fill the configuration file with the dataset path.",
        "default": "",
    },
    "template": {
        "help": "Use a template configuration file. Available are: id19_pag, id16_holo, id16_ctf, id16a_fluo, bm05_pag. You can also define your own templates via the NABU_TEMPLATES_PATH environment variable.",
        "default": "",
    },
    "helical": {"help": "Prepare configuration file for helical", "default": 0, "required": False, "type": int},
    "overwrite": {
        "help": "Whether to overwrite the output file if exists",
        "action": "store_const",
        "const": 1,
    },
}

# Default configuration for "zsplit" command
ZSplitConfig = {
    "input_file": {
        "help": "Input HDF5-Nexus file",
        "mandatory": True,
    },
    "output_directory": {
        "help": "Output directory to write split files.",
        "mandatory": True,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
    "entry": {
        "help": "HDF5 entry to take in the input file. By default, the first entry is taken.",
        "default": "",
    },
    "n_stages": {
        "help": "Number of expected stages (i.e different 'Z' values). By default it is inferred from the dataset.",
        "default": -1,
        "type": int,
    },
    "use_virtual_dataset": {
        "help": "Whether to use virtual datasets for output file. Not using a virtual dataset duplicates data and thus results in big files ! However virtual datasets currently have performance issues. Default is False",
        "default": 0,
        "type": int,
    },
}

# Default configuration for "histogram" command
HistogramConfig = {
    "h5_file": {
        "help": "HDF5 file(s). It can be one or several paths to HDF5 files. You can specify entry for each file with /path/to/file.h5?entry0000",
        "mandatory": True,
        "nargs": "+",
    },
    "output_file": {
        "help": "Output file (HDF5)",
        "mandatory": True,
    },
    "bins": {
        "help": "Number of bins for histogram if they have to be computed. Default is one million.",
        "default": 1000000,
        "type": int,
    },
    "chunk_size_slices": {
        "help": "If histogram are computed, specify the maximum subvolume size (in number of slices) for computing histogram.",
        "default": 100,
        "type": int,
    },
    "chunk_size_GB": {
        "help": "If histogram are computed, specify the maximum subvolume size (in GibaBytes) for computing histogram.",
        "default": -1,
        "type": float,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
}


# Default configuration for "reconstruct" command
ReconstructConfig = {
    "input_file": {
        "help": "Nabu input file",
        "default": "",
        "mandatory": True,
    },
    "logfile": {
        "help": "Log file. Default is dataset_prefix_nabu.log",
        "default": "",
    },
    "log_file": {
        "help": "Same as logfile. Deprecated, use --logfile instead.",
        "default": "",
    },
    "slice": {
        "help": "Slice(s) indice(s) to reconstruct, in the format z1-z2. Default (empty) is the whole volume. This overwrites the configuration file start_z and end_z. You can also use --slice first, --slice last, --slice middle, and --slice all",
        "default": "",
    },
    "gpu_mem_fraction": {
        "help": "Which fraction of GPU memory to use. Default is 0.9.",
        "default": 0.9,
        "type": float,
    },
    "cpu_mem_fraction": {
        "help": "Which fraction of memory to use. Default is 0.9.",
        "default": 0.9,
        "type": float,
    },
    "max_chunk_size": {
        "help": "Maximum chunk size to use.",
        "default": -1,
        "type": int,
    },
    "phase_margin": {
        "help": "Specify an explicit phase margin to use when performing phase retrieval.",
        "default": -1,
        "type": int,
    },
    "force_use_grouped_pipeline": {
        "help": "Force nabu to use the 'grouped' reconstruction pipeline - slower but should work for all big datasets.",
        "default": 0,
        "type": int,
    },
}


MultiCorConfig = ReconstructConfig.copy()
MultiCorConfig.update(
    {
        "cor": {
            "help": "Positions of the center of rotation. It must be a list of comma-separated scalars, or in the form start:stop:step, where start, stop and step can all be floating-point values.",
            "default": "",
            "mandatory": True,
        },
        "slice": {
            "help": "Slice(s) indice(s) to reconstruct, in the format z1-z2. Default (empty) is the whole volume. This overwrites the configuration file start_z and end_z. You can also use --slice first, --slice last, --slice middle, and --slice all",
            "default": "",
            "mandatory": True,
        },
    }
)


GenerateInfoConfig = {
    "hist_file": {
        "help": "HDF5 file containing the histogram, either the reconstruction file or a dedicated histogram file.",
        "default": "",
    },
    "hist_entry": {
        "help": "Histogram HDF5 entry. Defaults to the first available entry.",
        "default": "",
    },
    "output": {
        "help": "Output file name",
        "default": "",
        "mandatory": True,
    },
    "bliss_file": {
        "help": "HDF5 master file produced by BLISS",
        "default": "",
    },
    "bliss_entry": {
        "help": "Entry in the HDF5 master file produced by BLISS. By default, take the first entry.",
        "default": "",
    },
    "info_file": {
        "help": "Path to the .info file, in the case of a EDF dataset",
        "default": "",
    },
    "edf_proj": {
        "help": "Path to a projection, in the case of a EDF dataset",
        "default": "",
    },
}

RotateRadiosConfig = {
    "dataset": {
        "help": "Path to the dataset. Only HDF5 format is supported for now.",
        "default": "",
        "mandatory": True,
    },
    "entry": {
        "help": "HDF5 entry. By default, the first entry is taken.",
        "default": "",
    },
    "angle": {
        "help": "Rotation angle in degrees",
        "default": 0.0,
        "mandatory": True,
        "type": float,
    },
    "center": {
        "help": "Rotation center, in the form (x, y) where x (resp. y) is the horizontal (resp. vertical) dimension, i.e along the columns (resp. lines). Default is (Nx/2 - 0.5, Ny/2 - 0.5).",
        "default": "",
    },
    "output": {
        "help": "Path to the output file. Only HDF5 output is supported. In the case of HDF5 input, the output file will have the same structure.",
        "default": "",
        "mandatory": True,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
    "batchsize": {
        "help": "Size of the batch of images to process. Default is 100",
        "default": 100,
        "type": int,
    },
    "use_cuda": {
        "help": "Whether to use Cuda if available",
        "default": "1",
    },
    "use_multiprocessing": {
        "help": "Whether to use multiprocessing if available",
        "default": "1",
    },
}

DFFConfig = {
    "dataset": {
        "help": "Path to the dataset.",
        "default": "",
        "mandatory": True,
    },
    "entry": {
        "help": "HDF5 entry (for HDF5 datasets). By default, the first entry is taken.",
        "default": "",
    },
    "flatfield": {
        "help": "Whether to perform flat-field normalization. Default is True.",
        "default": "1",
        "type": int,
    },
    "sigma": {
        "default": 0.0,
        "help": "Enable high-pass filtering on double flatfield with this value of 'sigma'",
        "type": float,
    },
    "output": {
        "help": "Path to the output file (HDF5).",
        "default": "",
        "mandatory": True,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
    "chunk_size": {
        "help": "Maximum number of lines to read in each projection in a single pass. Default is 100",
        "default": 100,
        "type": int,
    },
}

CompareVolumesConfig = {
    "volume1": {
        "help": "Path to the first volume.",
        "default": "",
        "mandatory": True,
    },
    "volume2": {
        "help": "Path to the first volume.",
        "default": "",
        "mandatory": True,
    },
    "entry": {
        "help": "HDF5 entry. By default, the first entry is taken.",
        "default": "",
    },
    "hdf5_path": {
        "help": "Full HDF5 path to the data. Default is <entry>/reconstruction/results/data",
        "default": "",
    },
    "chunk_size": {
        "help": "Maximum number of images to read in each step. Default is 100.",
        "default": 100,
        "type": int,
    },
    "stop_at": {
        "help": "Stop the comparison immediately when the difference exceeds this threshold. Default is to compare the full volumes.",
        "default": "1e-4",
    },
    "statistics": {
        "help": "Compute statistics on the compared (sub-)volumes. Mind that in this case the command output will not be empty!",
        "default": 0,
        "type": int,
    },
}


# Default configuration for "stitching" command
StitchingConfig = {
    "input-file": {
        "help": "Nabu configuraiton file for stitching (can be obtain from nabu-stitching-boostrap command)",
        "default": "",
        "mandatory": True,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
    "--only-create-master-file": {
        "help": "Will create the master file with all sub files (volumes or scans). It expects the processing to be finished. It can happen if all slurm job have been submitted but you've been kicked out of the cluster of if you need to relaunch manually some failling job slurm for any reason",
        "default": False,
        "action": "store_true",
    },
}

# Default configuration for "stitching-bootstrap" command
BootstrapStitchingConfig = {
    "stitching-type": {
        "help": f"User can provide stitching type to filter some parameters. Must be in {StitchingType.values()}.",
        "default": None,
    },
    "level": {
        "help": "Level of options to embed in the configuration file. Can be 'required', 'optional', 'advanced'.",
        "default": "optional",
    },
    "output": {
        "help": "output file to store the configuration",
        "default": "stitching.conf",
    },
    "datasets": {
        "help": "datasets to be stitched together",
        "default": tuple(),
        "nargs": "*",
    },
}


ShrinkConfig = {
    "input_file": {
        "help": "Path to the NX file",
        "default": "",
        "mandatory": True,
    },
    "output_file": {
        "help": "Path to the output NX file",
        "default": "",
        "mandatory": True,
    },
    "entry": {
        "help": "HDF5 entry in the file. Default is to take the first entry.",
        "default": "",
    },
    "binning": {
        "help": "Binning factor, in the form (bin_z, bin_x). Each image (projection, dark, flat) will be binned by this factor",
        "default": "",
    },
    "subsampling": {"help": "Subsampling factor for projections (and metadata)", "default": ""},
    "threads": {
        "help": "Number of threads to use for binning. Default is 1.",
        "default": 1,
        "type": int,
    },
}

CompositeCorConfig = {
    "--filename_template": {
        "required": True,
        "help": """The filename template. It can optionally contain a segment equal to "X"*ndigits which will be replaced by the stage number if several stages are requested by the user""",
    },
    "--entry_name": {
        "required": False,
        "help": "Optional. The entry_name. It defaults to entry0000",
        "default": "entry0000",
    },
    "--num_of_stages": {
        "type": int,
        "required": False,
        "help": "Optional. How many stages. Example: from 0 to 43 -> --num_of_stages  44. It is optional. ",
    },
    "--oversampling": {
        "type": int,
        "default": 4,
        "required": False,
        "help": "Oversampling in the research of the axis position. Defaults to 4 ",
    },
    "--n_subsampling_y": {
        "type": int,
        "default": 10,
        "required": False,
        "help": "How many lines we are going to take from each radio. Defaults to 10.",
    },
    "--theta_interval": {
        "type": float,
        "default": 5,
        "required": False,
        "help": "Angular step for composing the image. Default to 5",
    },
    "--first_stage": {"type": int, "default": None, "required": False, "help": "Optional. The first stage.  "},
    "--output_file": {
        "type": str,
        "required": False,
        "help": "Optional. Where the list of cors will be written. Default is the filename postixed with cors.txt. If the output filename is postfixed with .json the output will be in json format",
    },
    "--cor_options": {
        "type": str,
        "help": """the cor_options string used by Nabu. Example 
        --cor_options "side='near'; near_pos = 300.0;  near_width = 20.0"
        """,
        "required": True,
    },
}

CreateDistortionMapHorizontallyMatchedFromPolyConfig = {
    "--nz": {"type": int, "help": "vertical dimension of the detector", "required": True},
    "--nx": {"type": int, "help": "horizontal dimension of the detector", "required": True},
    "--center_z": {"type": float, "help": "vertical position of the optical center", "required": True},
    "--center_x": {"type": float, "help": "horizontal position of the optical center", "required": True},
    "--c4": {"type": float, "help": "order 4 coefficient", "required": True},
    "--c2": {"type": float, "help": "order 2 coefficient", "required": True},
    "--target_file": {"type": str, "help": "The map output filename", "required": True},
    "--axis_pos": {
        "type": float,
        "default": None,
        "help": "Optional argument. If given it will be corrected for use with the produced map. The value is printed, or given as return argument if the utility is used from a script",
        "required": False,
    },
    "--loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
}

DiagToRotConfig = {
    "--diag_file": dict(
        required=True, help="The reconstruction file obtained by nabu-helical using the diag_zpro_run option", type=str
    ),
    "--entry_name": dict(
        required=False,
        help="entry_name. Defauls is entry0000",
        default="entry0000",
    ),
    "--near": dict(
        required=False,
        help="This is a relative offset respect to the center of the radios. The cor will be searched around the provided value. If not given the optinal parameter original_scan must be the original nexus file; and the estimated core will be taken there. The netry_name parameter also must be provided in this case",
        default=None,
        type=float,
    ),
    "--original_scan": dict(
        required=False,
        help="The original nexus file. Required only if near parameter is not given",
        default=None,
        type=str,
    ),
    "--entry_name": dict(
        required=False,
        help="The original nexus file entry name. Required only if near parameter is not given",
        default=None,
        type=str,
    ),
    "--near_width": dict(
        required=False,
        help="For the horizontal correlation, searching the cor. The radius around the near value",
        default=20,
        type=int,
    ),
    "--low_pass": dict(
        required=False,
        help="Data are filtered horizontally. details smaller than the provided value are filtered out. Default is 1( gaussian sigma)",
        default=1.0,
        type=float,
    ),
    "--high_pass": dict(
        required=False,
        help="Data are filtered horizontally. Bumps larger than the provided value are filtered out. Default is 10( gaussian sigma)",
        default=10,
        type=int,
    ),
    "--linear_interpolation": dict(
        required=False, help="If True(default) the cor will vary linearly with z_transl", default=True, type=str2bool
    ),
    "--use_l1_norm": dict(
        required=False,
        default=True,
        help="If false then a L2 norm will be used for the error metric, considering the overlaps, if true L1 norm will be considered",
        type=str2bool,
    ),
    "--cor_file": dict(required=True, help="The file where the information to correct the cor are written", type=str),
}

DiagToPixConfig = {
    "--diag_file": dict(
        required=True, help="The reconstruction file obtained by nabu-helical using the diag_zpro_run option", type=str
    ),
    "--entry_name": dict(required=False, help="entry_name. Defauls is entry0000", default="entry0000", type=str),
    "--search_radius_v": dict(
        required=False,
        help="For the vertical correlation, The maximal error in pixels of one turn respect to a contiguous one. Default is 20 ",
        default=20,
        type=int,
    ),
    "--nexus_target": dict(
        required=False,
        help="If given, the mentioned file will be edited with the proper pixel size, the proper COR, and corrected x_translations",
        default=None,
        type=str,
    ),
    "--nexus_source": dict(
        required=False,
        help="Optionaly given, used only if nexus_target has been give. The nexus file will be edited and written on nexus_target. Otherwise nexus_target is considered to be the source",
        default=None,
        type=str,
    ),
}

CorrectRotConfig = {
    "--cor_file": dict(required=True, help="The file produce by diag_to_rot", type=str),
    "--entry_name": dict(required=False, help="entry_name. Defauls is entry0000", default="entry0000", type=str),
    "--nexus_target": dict(
        required=True,
        help="The given file will be edited with the proper pixel size, the proper COR, and corrected x_translations",
        default=None,
        type=str,
    ),
    "--nexus_source": dict(
        required=True, help="The nexus file will be edited and written on nexus_target", default=None, type=str
    ),
}


ReduceDarkFlatConfig = {
    "dataset": {"help": "Dataset (NXtomo or EDF folder) to be treated", "mandatory": True},
    "entry": {
        "dest": "entry",
        "help": "an entry can be specify in case of an NXtomo",
        "default": None,
        "required": False,
    },
    "dark-method": {
        "help": f"Define the method to be used for computing darks. Valid methods are {ReduceMethod.values()}",
        "default": ReduceMethod.MEAN,
        "required": False,
    },
    "flat-method": {
        "help": f"Define the method to be used for computing flats. Valid methods are {ReduceMethod.values()}",
        "default": ReduceMethod.MEDIAN,
        "required": False,
    },
    "overwrite": {
        "dest": "overwrite",
        "action": "store_true",
        "default": False,
        "help": "Overwrite dark/flats if exists",
        "required": False,
    },
    "debug": {
        "dest": "debug",
        "action": "store_true",
        "default": False,
        "help": "Set logging system in debug mode",
        "required": False,
    },
    "output-reduced-flats-file": {
        "aliases": ("orfl",),
        "default": None,
        "help": "Where to save reduced flats. If not provided will be dump near the .nx file at {scan_prefix}_flats.hdf5",
        "required": False,
    },
    "output-reduced-flats-data-path": {
        "aliases": ("output-reduced-flats-dp", "orfdp"),
        "default": None,
        "help": "Path in the output reduced flats file to save the dataset. If not provided will be saved at {entry}/flats/",
        "required": False,
    },
    "output-reduced-darks-file": {
        "aliases": ("ordf",),
        "default": None,
        "help": "Where to save reduced dark. If not provided will be dump near the .nx file at {scan_prefix}_darks.hdf5",
        "required": False,
    },
    "output-reduced-darks-data-path": {
        "aliases": ("output-reduced-darks-dp", "orddp"),
        "default": None,
        "help": "Path in the output reduced darks file to save the dataset. If not provided will be saved at {entry}/darks/",
        "required": False,
    },
}

ShowReconstructionTimingsConfig = {
    "logfile": {
        "help": "Path to the log file.",
        "default": "",
        "mandatory": True,
    },
    "cutoff": {
        "help": "Cut-off parameter. Timings below this value will be discarded. For a upper-bound cutoff, provide a value in the form 'low, high'",
        "default": None,
    },
    "type": {
        "help": "How to display the result. Default is a pie chart. Possible values are: pie, bars, violin",
        "default": "pie",
        "type": str,
    },
}



import numpy as np
import pycuda.driver as cuda
from ..utils import get_cuda_srcfile, check_supported, docstring
from ..cuda.processing import CudaProcessing
from ..processing.fft_cuda import get_fft_class
from .phase import PaganinPhaseRetrieval


class CudaPaganinPhaseRetrieval(PaganinPhaseRetrieval):
    supported_paddings = ["zeros", "constant", "edge"]

    @docstring(PaganinPhaseRetrieval)
    def __init__(
        self,
        shape,
        distance=0.5,
        energy=20,
        delta_beta=250.0,
        pixel_size=1e-6,
        padding="edge",
        cuda_options=None,
        fftw_num_threads=None,  # COMPAT.
        fft_num_threads=None,
        fft_backend="vkfft",
    ):
        """
        Please refer to the documentation of
        nabu.preproc.phase.PaganinPhaseRetrieval
        """
        padding = self._check_padding(padding)
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        super().__init__(
            shape,
            distance=distance,
            energy=energy,
            delta_beta=delta_beta,
            pixel_size=pixel_size,
            padding=padding,
            use_rfft=True,
            fft_num_threads=False,
        )
        self._init_gpu_arrays()
        self._init_fft(fft_backend)
        self._init_padding_kernel()
        self._init_mult_kernel()

    def _check_padding(self, padding):
        check_supported(padding, self.supported_paddings, "padding")
        if padding == "zeros":
            padding = "constant"
        return padding

    def _init_gpu_arrays(self):
        self.d_paganin_filter = self.cuda_processing.to_device(
            "d_paganin_filter", np.ascontiguousarray(self.paganin_filter, dtype=np.float32)
        )

    # overwrite parent method, don't initialize any FFT plan
    def _get_fft(self, use_rfft, fft_num_threads):
        self.use_rfft = use_rfft

    def _init_fft(self, fft_backend):
        fft_cls = get_fft_class(backend=fft_backend)
        self.cufft = fft_cls(shape=self.data_padded.shape, dtype=np.float32, r2c=True)
        self.d_radio_padded = self.cuda_processing.allocate_array("d_radio_padded", self.cufft.shape, "f")
        self.d_radio_f = self.cuda_processing.allocate_array("d_radio_f", self.cufft.shape_out, np.complex64)

    def _init_padding_kernel(self):
        kern_signature = {"constant": "Piiiiiiiiffff", "edge": "Piiiiiiii"}
        self.padding_kernel = self.cuda_processing.kernel(
            "padding_%s" % self.padding,
            filename=get_cuda_srcfile("padding.cu"),
            signature=kern_signature[self.padding],
        )
        Ny, Nx = self.shape
        Nyp, Nxp = self.shape_padded
        self.padding_kernel_args = [
            self.d_radio_padded,
            Nx,
            Ny,
            Nxp,
            Nyp,
            self.pad_left_len,
            self.pad_right_len,
            self.pad_top_len,
            self.pad_bottom_len,
        ]
        # TODO configurable constant values
        if self.padding == "constant":
            self.padding_kernel_args.extend([0, 0, 0, 0])

    def _init_mult_kernel(self):
        self.cpxmult_kernel = self.cuda_processing.kernel(
            "inplace_complexreal_mul_2Dby2D",
            filename=get_cuda_srcfile("ElementOp.cu"),
            signature="PPii",
        )
        self.cpxmult_kernel_args = [
            self.d_radio_f,
            self.d_paganin_filter,
            self.shape_padded[1] // 2 + 1,
            self.shape_padded[0],
        ]

    def set_input(self, data):
        assert data.shape == self.shape
        assert data.dtype == np.float32
        # Rectangular memcopy
        # TODO profile, and if needed include this copy in the padding kernel
        if isinstance(data, np.ndarray) or isinstance(data, self.cuda_processing.array_class):
            self.d_radio_padded[: self.shape[0], : self.shape[1]] = data[:, :]
        elif isinstance(data, cuda.DeviceAllocation):
            # TODO manual memcpy2D
            raise NotImplementedError("pycuda buffers are not supported yet")
        else:
            raise ValueError("Expected either numpy array, pycuda array or pycuda buffer")

    def get_output(self, output):
        s0, s1 = self.shape
        if output is None:
            # copy D2H
            return self.d_radio_padded[:s0, :s1].get()
        assert output.shape == self.shape
        assert output.dtype == np.float32
        output[:, :] = self.d_radio_padded[:s0, :s1]
        return output

    def apply_filter(self, radio, output=None):
        self.set_input(radio)

        self.padding_kernel(*self.padding_kernel_args)
        self.cufft.fft(self.d_radio_padded, output=self.d_radio_f)
        self.cpxmult_kernel(*self.cpxmult_kernel_args)
        self.cufft.ifft(self.d_radio_f, output=self.d_radio_padded)

        return self.get_output(output)

    __call__ = apply_filter

    retrieve_phase = apply_filter



from tomoscan.identifier import BaseIdentifier
from nabu.stitching.stitcher.y_stitcher import PreProcessingYStitcher as PreProcessYStitcher
from nabu.stitching.config import PreProcessedYStitchingConfiguration


def y_stitching(configuration: PreProcessedYStitchingConfiguration, progress=None) -> BaseIdentifier:
    """
    Apply stitching from provided configuration.
    Stitching will be applied along the first axis - 1 (aka y).

    like:
                    axis 0
                      ^
                      |
    x-ray             |
    -------->          ------> axis 2
                     /
                    /
                     axis 1
    """
    if isinstance(configuration, PreProcessedYStitchingConfiguration):
        stitcher = PreProcessYStitcher(configuration=configuration, progress=progress)
    else:
        raise TypeError(
            f"configuration is expected to be in {(PreProcessedYStitchingConfiguration, )}. {type(configuration)} provided"
        )
    return stitcher.stitch()



import numpy as np
from ..utils import get_opencl_srcfile
from ..opencl.processing import OpenCLProcessing
from .padding_base import PaddingBase
from ..opencl.utils import __has_pyopencl__

if __has_pyopencl__:
    from ..opencl.memcpy import OpenCLMemcpy2D


class OpenCLPadding(PaddingBase):
    """
    A class for performing padding on GPU using OpenCL
    """

    backend = "opencl"

    # TODO docstring from base class
    def __init__(self, shape, pad_width, mode="constant", opencl_options=None, **kwargs):
        super().__init__(shape, pad_width, mode=mode, **kwargs)
        self.opencl_processing = self.processing = OpenCLProcessing(**(opencl_options or {}))
        self.queue = self.opencl_processing.queue
        self._init_opencl_coordinate_transform()

    def _init_opencl_coordinate_transform(self):
        if self.mode == "constant":
            self.d_padded_array_constant = self.processing.to_device(
                "d_padded_array_constant", self.padded_array_constant
            )
            self.memcpy2D = OpenCLMemcpy2D(ctx=self.processing.ctx, queue=self.queue)  # pylint: disable=E0606
            return
        self._coords_transform_kernel = self.processing.kernel(
            "coordinate_transform",
            filename=get_opencl_srcfile("padding.cl"),
        )
        self._coords_transform_global_size = self.padded_shape[::-1]
        self.d_coords_rows = self.processing.to_device("d_coords_rows", self.coords_rows)
        self.d_coords_cols = self.processing.to_device("d_coords_cols", self.coords_cols)

    def _pad_constant(self, image, output):
        pad_y, pad_x = self.pad_width
        # the following line is not implemented in pyopencl
        # self.d_padded_array_constant[pad_y[0] : pad_y[0] + self.shape[0], pad_x[0] : pad_x[0] + self.shape[1]] = image[:]
        # cl.enqueue_copy is too cumbersome to use for Buffer <-> Buffer.
        # Use a dedicated kernel instead.
        # This is not optimal (two copies) - TODO write a constant padding kernel
        self.memcpy2D(self.d_padded_array_constant, image, image.shape[::-1], dst_offset_xy=(pad_x[0], pad_y[0]))
        output[:] = self.d_padded_array_constant[:]
        return output

    def pad(self, image, output=None):
        """
        Pad an array.

        Parameters
        ----------
        image: pyopencl array
            Image to pad
        output: pyopencl array
            Output image. If provided, must be in the expected shape.
        """
        if output is None:
            output = self.processing.allocate_array("d_output", self.padded_shape)
        if self.mode == "constant":
            return self._pad_constant(image, output)
        self._coords_transform_kernel(
            self.queue,
            image,
            output,
            self.d_coords_cols,
            self.d_coords_rows,
            np.int32(self.shape[1]),
            np.int32(self.padded_shape[1]),
            np.int32(self.padded_shape[0]),
            global_size=self._coords_transform_global_size,
        )
        return output



import h5py
import numpy
import logging
from typing import Optional
from .base import DumperBase
from nabu.stitching.config import PostProcessedSingleAxisStitchingConfiguration
from nabu import version as nabu_version
from nabu.io.writer import get_datetime
from tomoscan.identifier import VolumeIdentifier
from tomoscan.volumebase import VolumeBase
from tomoscan.esrf.volume import HDF5Volume
from tomoscan.io import HDF5File
from contextlib import AbstractContextManager


_logger = logging.getLogger(__name__)


class OutputVolumeContext(AbstractContextManager):
    """
    Utils class to Manage the data volume creation and save it (data only !). target: used for volume stitching
    In the case of HDF5 we want to save this directly in the file to avoid
    keeping the full volume in memory.
    Insure also contain processing will be common between the different processing

    If stitching_sources_arr_shapes is provided this mean that we want to create stitching region and then create a VDS to avoid data duplication
    """

    def __init__(
        self,
        volume: VolumeBase,
        volume_shape: tuple,
        dtype: numpy.dtype,
        dumper,
    ) -> None:
        super().__init__()
        if not isinstance(volume, VolumeBase):
            raise TypeError(f"Volume is expected to be an instance of {VolumeBase}. {type(volume)} provided instead")

        self._volume = volume
        self._volume_shape = volume_shape
        self.__file_handler = None
        self._dtype = dtype
        self._dumper = dumper

    @property
    def _file_handler(self):
        return self.__file_handler

    def _build_hdf5_output(self):
        return self._file_handler.create_dataset(
            self._volume.data_url.data_path(),
            shape=self._volume_shape,
            dtype=self._dtype,
        )

    def _create_stitched_volume_dataset(self):
        # handle the specific case of HDF5. Goal: avoid getting the full stitched volume in memory
        if isinstance(self._volume, HDF5Volume):
            self.__file_handler = HDF5File(self._volume.data_url.file_path(), mode="a")
            # if need to delete an existing dataset
            if self._volume.overwrite and self._volume.data_path in self._file_handler:
                try:
                    del self._file_handler[self._volume.data_path]
                except Exception as e:
                    _logger.error(f"Fail to overwrite data. Reason is {e}")
                    data = None
                    self._file_handler.close()
                    self._duplicate_data = True
                    # avoid creating a dataset for stitched volume as creation of the stitched_volume failed
                    return data

            # create dataset
            try:
                data = self._build_hdf5_output()
            except Exception as e2:
                _logger.error(f"Fail to create final dataset. Reason is {e2}")
                data = None
                self._file_handler.close()
                self._duplicate_data = True
                # avoid creating a dataset for stitched volume as creation of the stitched_volume failed
        else:
            raise TypeError("only HDF5 output is handled")
        # else:
        #     # for other file format: create the full dataset in memory before dumping it
        #     data = numpy.empty(self._volume_shape, dtype=self._dtype)
        #     self._volume.data = data
        return data

    def __enter__(self):
        assert self._dumper.output_dataset is None
        self._dumper.output_dataset = self._create_stitched_volume_dataset()
        return self._dumper.output_dataset

    def __exit__(self, exc_type, exc_value, traceback):
        if self._file_handler is not None:
            return self._file_handler.close()
        else:
            self._volume.save_data()


class OutputVolumeNoDDContext(OutputVolumeContext):
    """
    Dedicated output volume context for saving a volume without Data Duplication (DD)
    """

    def __init__(
        self,
        volume: VolumeBase,
        volume_shape: tuple,
        dtype: numpy.dtype,
        dumper,
        stitching_sources_arr_shapes: Optional[tuple],
    ) -> None:
        if not isinstance(dumper, PostProcessingStitchingDumperNoDD):
            raise TypeError
        # TODO: compute volume_shape from here
        self._stitching_sources_arr_shapes = stitching_sources_arr_shapes

        super().__init__(volume, volume_shape, dtype, dumper)

    def __enter__(self):
        dataset = super().__enter__()
        assert isinstance(self._dumper, PostProcessingStitchingDumperNoDD)
        self._dumper.stitching_regions_hdf5_dataset = self._create_stitched_sub_region_datasets()
        return dataset

    def _build_hdf5_output(self):
        return h5py.VirtualLayout(
            shape=self._volume_shape,
            dtype=self._dtype,
        )

    def __exit__(self, exc_type, exc_value, traceback):
        # in the case of no data duplication we need to create the virtual dataset at the end
        if not isinstance(self._dumper.output_dataset, h5py.VirtualLayout):
            raise TypeError("dumper output_dataset should be a virtual layout")
        self._file_handler.create_virtual_dataset(self._volume.data_url.data_path(), layout=self._dumper.output_dataset)
        super().__exit__(exc_type=exc_type, exc_value=exc_value, traceback=traceback)

    def _create_stitched_sub_region_datasets(self):
        # create datasets to store overlaps if needed
        if not isinstance(self._volume, HDF5Volume):
            raise TypeError("Avoid Data Duplication is only available for HDF5 output volume")

        stitching_regions_hdf5_dataset = []
        for i_region, overlap_shape in enumerate(self._stitching_sources_arr_shapes):
            data_path = f"{self._volume.data_path}/stitching_regions/region_{i_region}"
            if self._volume.overwrite and data_path in self._file_handler:
                del self._file_handler[data_path]
            stitching_regions_hdf5_dataset.append(
                self._file_handler.create_dataset(
                    name=data_path,
                    shape=overlap_shape,
                    dtype=self._dtype,
                )
            )
        self._dumper.stitching_regions_hdf5_dataset = stitching_regions_hdf5_dataset
        return stitching_regions_hdf5_dataset


class PostProcessingStitchingDumper(DumperBase):
    """
    dumper to be used when save data durint post-processing stitching (on recosntructed volume). Output is expected to be an NXtomo
    """

    OutputDatasetContext = OutputVolumeContext

    def __init__(self, configuration) -> None:
        if not isinstance(configuration, PostProcessedSingleAxisStitchingConfiguration):
            raise TypeError(
                f"configuration is expected to be an instance of {PostProcessedSingleAxisStitchingConfiguration}. Get {type(configuration)} instead"
            )
        super().__init__(configuration)
        self._output_dataset = None
        self._input_volumes = configuration.input_volumes

    def save_configuration(self):
        voxel_size = self._input_volumes[0].voxel_size

        def get_position():
            # the z-serie is z-ordered from higher to lower. We can reuse this with pixel size and shape to
            # compute the position of the stitched volume
            if voxel_size is None:
                return None
            return numpy.array(self._input_volumes[0].position) + voxel_size * (
                numpy.array(self._input_volumes[0].get_volume_shape()) / 2.0
                - numpy.array(self.configuration.output_volume.get_volume_shape()) / 2.0
            )

        self.configuration.output_volume.voxel_size = voxel_size or ""
        try:
            self.configuration.output_volume.position = get_position()
        except Exception:
            self.configuration.output_volume.position = numpy.array([0, 0, 0])

        self.configuration.output_volume.metadata.update(
            {
                "about": {
                    "program": "nabu-stitching",
                    "version": nabu_version,
                    "date": get_datetime(),
                },
                "configuration": self.configuration.to_dict(),
            }
        )
        self.configuration.output_volume.save_metadata()

    @property
    def output_identifier(self) -> VolumeIdentifier:
        return self.configuration.output_volume.get_identifier()

    def create_output_dataset(self):
        """
        function called at the beginning of the stitching to prepare output dataset
        """
        self._dataset = h5py.VirtualLayout(
            shape=self._volume_shape,
            dtype=self._dtype,
        )


class PostProcessingStitchingDumperNoDD(PostProcessingStitchingDumper):
    """
    same as PostProcessingStitchingDumper but prevent to do data duplication.
    In this case we need to work on HDF5 file only
    """

    OutputDatasetContext = OutputVolumeNoDDContext

    def __init__(self, configuration) -> None:
        if not isinstance(configuration, PostProcessedSingleAxisStitchingConfiguration):
            raise TypeError(
                f"configuration is expected to be an instance of {PostProcessedSingleAxisStitchingConfiguration}. Get {type(configuration)} instead"
            )
        super().__init__(configuration)
        self._stitching_regions_hdf5_dataset = None
        self._raw_regions_hdf5_dataset = None

    def create_output_dataset(self):
        """
        function called at the beginning of the stitching to prepare output dataset
        """
        self._dataset = h5py.VirtualLayout(
            shape=self._volume_shape,
            dtype=self._dtype,
        )

    @staticmethod
    def create_subset_selection(dataset: h5py.Dataset, slices: tuple) -> h5py.VirtualSource:
        assert isinstance(dataset, h5py.Dataset), f"dataset is expected to be a h5py.Dataset. Get {type(dataset)}"
        assert isinstance(slices, tuple), f"slices is expected to be a tuple of slices. Get {type(slices)} instead"
        import h5py._hl.selections as selection

        virtual_source = h5py.VirtualSource(dataset)
        sel = selection.select(dataset.shape, slices, dataset=dataset)
        virtual_source.sel = sel
        return virtual_source

    @PostProcessingStitchingDumper.output_dataset.setter
    def output_dataset(self, dataset: Optional[h5py.VirtualLayout]):
        if dataset is not None and not isinstance(dataset, h5py.VirtualLayout):
            raise TypeError("in the case we want to avoid data duplication 'output_dataset' must be a VirtualLayout")
        self._output_dataset = dataset

    @property
    def stitching_regions_hdf5_dataset(self) -> Optional[tuple]:
        """hdf5 dataset storing the stitched regions"""
        return self._stitching_regions_hdf5_dataset

    @stitching_regions_hdf5_dataset.setter
    def stitching_regions_hdf5_dataset(self, datasets: tuple):
        self._stitching_regions_hdf5_dataset = datasets

    @property
    def raw_regions_hdf5_dataset(self) -> Optional[tuple]:
        """hdf5 raw dataset"""
        return self._raw_regions_hdf5_dataset

    @raw_regions_hdf5_dataset.setter
    def raw_regions_hdf5_dataset(self, datasets: tuple):
        self._raw_regions_hdf5_dataset = datasets

    def save_stitched_frame(
        self,
        stitched_frame: numpy.ndarray,
        composition_cls: dict,
        i_frame: int,
        axis: int,
    ):
        """
        Save the full stitched frame to disk
        """
        output_dataset = self.output_dataset
        if output_dataset is None:
            raise ValueError("output_dataset must be set before calling any frame stitching")
        stitching_regions_hdf5_dataset = self.stitching_regions_hdf5_dataset
        if stitching_regions_hdf5_dataset is None:
            raise ValueError("stitching_region_hdf5_dataset must be set before calling any frame stitching")
        raw_regions_hdf5_dataset = self.raw_regions_hdf5_dataset

        # save stitched region
        stitching_regions = composition_cls["overlap_composition"]
        for (_, _, region_start, region_end), stitching_region_hdf5_dataset in zip(
            stitching_regions.browse(), stitching_regions_hdf5_dataset
        ):
            assert isinstance(output_dataset, h5py.VirtualLayout)
            assert isinstance(stitching_region_hdf5_dataset, h5py.Dataset)
            stitching_region_array = stitched_frame[region_start:region_end]
            self.save_frame_to_disk(
                output_dataset=stitching_region_hdf5_dataset,
                index=i_frame,
                stitched_frame=stitching_region_array,
                axis=1,
                region_start=0,
                region_end=None,
            )
            vs = self.create_subset_selection(
                dataset=stitching_region_hdf5_dataset,
                slices=(
                    slice(0, stitching_region_hdf5_dataset.shape[0]),
                    slice(i_frame, i_frame + 1),
                    slice(0, stitching_region_hdf5_dataset.shape[2]),
                ),
            )

            self.save_frame_to_disk(
                output_dataset=output_dataset,
                index=i_frame,
                axis=axis,
                region_start=region_start,
                region_end=region_end,
                stitched_frame=vs,
            )

        # create virtual source of the raw data
        raw_regions = composition_cls["raw_composition"]
        for (frame_start, frame_end, region_start, region_end), raw_region_hdf5_dataset in zip(
            raw_regions.browse(), raw_regions_hdf5_dataset
        ):
            vs = self.create_subset_selection(
                dataset=raw_region_hdf5_dataset,
                slices=(
                    slice(frame_start, frame_end),
                    slice(i_frame, i_frame + 1),
                    slice(0, raw_region_hdf5_dataset.shape[2]),
                ),
            )
            self.save_frame_to_disk(
                output_dataset=output_dataset,
                index=i_frame,
                axis=1,
                region_start=region_start,
                region_end=region_end,
                stitched_frame=vs,
            )



import numpy as np
import pycuda.driver as cuda
from ..utils import updiv, get_cuda_srcfile
from ..cuda.utils import copy_array, check_textures_availability
from ..cuda.processing import CudaProcessing
from ..cuda.kernel import CudaKernel
from .filtering_cuda import CudaSinoFilter
from .sinogram_cuda import CudaSinoMult
from .fbp_base import BackprojectorBase


class CudaBackprojector(BackprojectorBase):
    backend = "cuda"
    kernel_filename = "backproj.cu"
    backend_processing_class = CudaProcessing
    SinoFilterClass = CudaSinoFilter
    SinoMultClass = CudaSinoMult

    def _check_textures_availability(self):
        self._use_textures = self.extra_options.get("use_textures", True) and check_textures_availability()

    def _get_kernel_signature(self):
        kern_full_sig = list("PPiifiiffPPPf")
        if self._axis_correction is None:
            kern_full_sig[11] = ""
        if self._use_textures:
            # texture references - no object is passed (deprecated, removed in Cuda 12)
            kern_full_sig[1] = ""
        return "".join(kern_full_sig)

    def _get_kernel_options(self):
        super()._get_kernel_options()
        self._kernel_options.update(
            {
                "file_name": get_cuda_srcfile(self.kernel_filename),
                "kernel_signature": self._get_kernel_signature(),
                "texture_name": "tex_projections",
            }
        )

    def _prepare_kernel_args(self):
        super()._prepare_kernel_args()
        self.kern_proj_kwargs.update(
            {
                "shared_size": self._kernel_options["shared_size"],
            }
        )
        # texture references - no object is passed (deprecated, removed in Cuda 12)
        if self._use_textures:
            self.kern_proj_args.pop(1)
        else:
            self._d_sino = self._processing.allocate_array("_d_sino", self.sino_shape)
            self.kern_proj_args[1] = self._d_sino.gpudata

    def _prepare_textures(self):
        if self._use_textures:
            self.texref_proj = self.gpu_projector.module.get_texref(self._kernel_options["texture_name"])
            self.texref_proj.set_filter_mode(cuda.filter_mode.LINEAR)
            self.gpu_projector.prepare(self._kernel_options["kernel_signature"], [self.texref_proj])
            # Bind texture
            self._d_sino_cua = cuda.np_to_array(np.zeros(self.sino_shape, "f"), "C")
            self.texref_proj.set_array(self._d_sino_cua)
        else:
            # d_sino_ref = self._d_sino.gpudata
            # self.kern_proj_args.insert(2, d_sino_ref)
            self.gpu_projector.prepare(self._kernel_options["kernel_signature"], [])

    def _compile_kernels(self):
        self._prepare_kernel_args()
        if self._use_textures:
            self._kernel_options["sourcemodule_options"].append("-DUSE_TEXTURES")
        self.gpu_projector = CudaKernel(
            self._kernel_options["kernel_name"],
            filename=self._kernel_options["file_name"],
            options=self._kernel_options["sourcemodule_options"],
            silent_compilation_warnings=True,  # textures and Cuda 11
        )
        if self.halftomo and self.rot_center < self.dwidth:
            self.sino_mult = CudaSinoMult(self.sino_shape, self.rot_center, ctx=self._processing.ctx)
        self._prepare_textures()  # has to be done after compilation for Cuda (to bind texture to built kernel)

    def _transfer_to_texture(self, sino, do_checks=True):
        if do_checks and not (sino.flags.c_contiguous):
            raise ValueError("Expected C-Contiguous array")
        if self._use_textures:
            copy_array(self._d_sino_cua, sino, check=do_checks)
        else:
            if id(self._d_sino) == id(sino):
                return
            self._d_sino[:] = sino[:]


# COMPAT.
Backprojector = CudaBackprojector


class PolarBackprojector(Backprojector):
    """
    Cuda Backprojector with output in polar coordinates.
    """

    cuda_fname = "backproj_polar.cu"
    cuda_kernel_name = "backproj_polar"

    # patch parent method: force slice_shape to (n_angles, n_x)
    def _set_angles(self, angles, n_angles):
        Backprojector._set_angles(self, angles, n_angles)
        self.slice_shape = (self.n_angles, self.n_x)

    # patch parent method:
    def _set_slice_roi(self, slice_roi):
        if slice_roi is not None:
            raise ValueError("slice_roi is not supported with this class")
        Backprojector._set_slice_roi(self, slice_roi)

    # patch parent method: don't do the 4X compute-workload optimization for this kernel
    def _get_kernel_options(self):
        Backprojector._get_kernel_options(self)
        block = self._kernel_options["block"]
        self._kernel_options["grid"] = (updiv(self.n_x, block[0]), updiv(self.n_y, block[1]))

    # patch parent method: update kernel args
    def _compile_kernels(self):
        n_y = self.n_y
        self.n_y = self.n_angles
        Backprojector._compile_kernels(self)
        self.n_y = n_y



import numpy
import logging
from typing import Optional, Union
from silx.utils.enum import Enum as _Enum
from nabu.misc import fourier_filters
from scipy.fft import rfftn as local_fftn
from scipy.fft import irfftn as local_ifftn
from tomoscan.utils.geometry import BoundingBox1D

_logger = logging.getLogger(__name__)


class OverlapStitchingStrategy(_Enum):
    MEAN = "mean"
    COSINUS_WEIGHTS = "cosinus weights"
    LINEAR_WEIGHTS = "linear weights"
    CLOSEST = "closest"
    IMAGE_MINIMUM_DIVERGENCE = "image minimum divergence"
    HIGHER_SIGNAL = "higher signal"


DEFAULT_OVERLAP_STRATEGY = OverlapStitchingStrategy.COSINUS_WEIGHTS

DEFAULT_OVERLAP_SIZE = None  # could also be an int
# default overlap to be take for stitching. Ig None: take the largest possible area


class OverlapKernelBase:
    pass


class ImageStichOverlapKernel(OverlapKernelBase):
    """
    Stitch two images along Y (axis 0 in image space)
    """

    DEFAULT_HIGH_FREQUENCY_THRESHOLD = 2

    def __init__(
        self,
        stitching_axis: int,
        frame_unstitched_axis_size: tuple,
        stitching_strategy: OverlapStitchingStrategy = DEFAULT_OVERLAP_STRATEGY,
        overlap_size: int = DEFAULT_OVERLAP_SIZE,
        extra_params: Optional[dict] = None,
    ) -> None:
        """
        :param stitching_axis: axis along which stitching is operate. Must be in '0', '1'
        :param frame_unstitched_axis_size: according to the stitching axis the stitched framed will always have a constant size:
            * If stitching_axis == 0 then it will be the frame width
            * If stitching_axis == 1 then it will be the frame height
        :param stitching_strategy: stategy / algorithm to use in order to generate the stitching
        :param overlap_size: size (int) of the overlap (stitching) between the two images
        :param extra_params: possibly extra parameters to operate the stitching
        """
        from nabu.stitching.config import KEY_THRESHOLD_FREQUENCY  # avoid acylic import

        if not isinstance(overlap_size, int) and overlap_size > 0:
            raise TypeError(
                f"overlap_size is expected to be a positive int, {overlap_size} - not {overlap_size} ({type(overlap_size)})"
            )
        if not isinstance(frame_unstitched_axis_size, int) or not frame_unstitched_axis_size > 0:
            raise TypeError(
                f"frame_width is expected to be a positive int, {frame_unstitched_axis_size} - not {frame_unstitched_axis_size} ({type(frame_unstitched_axis_size)})"
            )

        if not stitching_axis in (0, 1):
            raise ValueError(
                "stitching_axis is expected to be the axis along which stitching must be done. It should be '0' or '1'"
            )

        self._stitching_axis = stitching_axis
        self._overlap_size = abs(overlap_size)
        self._frame_unstitched_axis_size = frame_unstitched_axis_size
        self._stitching_strategy = OverlapStitchingStrategy.from_value(stitching_strategy)
        self._weights_img_1 = None
        self._weights_img_2 = None
        if extra_params is None:
            extra_params = {}
        self._high_frequency_threshold = extra_params.get(
            KEY_THRESHOLD_FREQUENCY, self.DEFAULT_HIGH_FREQUENCY_THRESHOLD
        )

    def __str__(self) -> str:
        return f"z-stitching kernel (policy={self.stitching_strategy.value}, overlap_size={self.overlap_size}, frame={self._frame_unstitched_axis_size})"

    @staticmethod
    def __check_img(img, name):
        if not isinstance(img, numpy.ndarray) and img.ndim == 2:
            raise ValueError(f"{name} is expected to be 2D numpy array")

    @property
    def stitched_axis(self) -> int:
        return self._stitching_axis

    @property
    def unstitched_axis(self) -> int:
        """
        util function. The kernel is operating stitching on images along a single axis (`stitching_axis`).
        This property is returning the other axis.
        """
        if self.stitched_axis == 0:
            return 1
        else:
            return 0

    @property
    def overlap_size(self) -> int:
        return self._overlap_size

    @overlap_size.setter
    def overlap_size(self, size: int):
        if not isinstance(size, int):
            raise TypeError(f"height expects a int ({type(size)} provided instead)")
        if not size >= 0:
            raise ValueError(f"height is expected to be positive")
        self._overlap_size = abs(size)
        # update weights if needed
        if self._weights_img_1 is not None or self._weights_img_2 is not None:
            self.compute_weights()

    @property
    def img_2(self) -> numpy.ndarray:
        return self._img_2

    @property
    def weights_img_1(self) -> Optional[numpy.ndarray]:
        return self._weights_img_1

    @property
    def weights_img_2(self) -> Optional[numpy.ndarray]:
        return self._weights_img_2

    @property
    def stitching_strategy(self) -> OverlapStitchingStrategy:
        return self._stitching_strategy

    def compute_weights(self):
        if self.stitching_strategy is OverlapStitchingStrategy.MEAN:
            weights_img_1 = numpy.ones(self._overlap_size) * 0.5
            weights_img_2 = weights_img_1[::-1]
        elif self.stitching_strategy is OverlapStitchingStrategy.CLOSEST:
            n_item = self._overlap_size // 2 + self._overlap_size % 2
            weights_img_1 = numpy.concatenate(
                [
                    numpy.ones(n_item),
                    numpy.zeros(self._overlap_size - n_item),
                ]
            )
            weights_img_2 = numpy.concatenate(
                [
                    numpy.zeros(n_item),
                    numpy.ones(self._overlap_size - n_item),
                ]
            )
        elif self.stitching_strategy is OverlapStitchingStrategy.LINEAR_WEIGHTS:
            weights_img_1 = numpy.linspace(1.0, 0.0, self._overlap_size)
            weights_img_2 = weights_img_1[::-1]
        elif self.stitching_strategy is OverlapStitchingStrategy.COSINUS_WEIGHTS:
            angles = numpy.linspace(0.0, numpy.pi / 2.0, self._overlap_size)
            weights_img_1 = numpy.cos(angles) ** 2
            weights_img_2 = numpy.sin(angles) ** 2
        elif self.stitching_strategy in (
            OverlapStitchingStrategy.IMAGE_MINIMUM_DIVERGENCE,
            OverlapStitchingStrategy.HIGHER_SIGNAL,
        ):
            # those strategies are not using constant weights but have treatments depending on the provided img_1 and mg_2 during stitching
            return
        else:
            raise NotImplementedError(f"{self.stitching_strategy} not implemented")

        if self._stitching_axis == 0:
            self._weights_img_1 = weights_img_1.reshape(-1, 1) * numpy.ones(self._frame_unstitched_axis_size).reshape(
                1, -1
            )
            self._weights_img_2 = weights_img_2.reshape(-1, 1) * numpy.ones(self._frame_unstitched_axis_size).reshape(
                1, -1
            )
        elif self._stitching_axis == 1:
            self._weights_img_1 = weights_img_1.reshape(1, -1) * numpy.ones(self._frame_unstitched_axis_size).reshape(
                -1, 1
            )
            self._weights_img_2 = weights_img_2.reshape(1, -1) * numpy.ones(self._frame_unstitched_axis_size).reshape(
                -1, 1
            )
        else:
            raise ValueError(f"stitching_axis should be in (0, 1). {self._stitching_axis} provided")

    def stitch(self, img_1, img_2, check_input=True) -> tuple:
        """Compute overlap region from the defined strategy"""
        if check_input:
            self.__check_img(img_1, "img_1")
            self.__check_img(img_2, "img_2")

            if img_1.shape != img_2.shape:
                raise ValueError(
                    f"both images are expected to be of the same shape to apply stitch ({img_1.shape} vs {img_2.shape})"
                )

        if self._stitching_strategy is OverlapStitchingStrategy.IMAGE_MINIMUM_DIVERGENCE:
            return (
                compute_image_minimum_divergence(
                    img_1=img_1,
                    img_2=img_2,
                    high_frequency_threshold=self._high_frequency_threshold,
                    stitching_axis=self.stitched_axis,
                ),
                None,
                None,
            )
        elif self._stitching_strategy is OverlapStitchingStrategy.HIGHER_SIGNAL:
            return (
                compute_image_higher_signal(
                    img_1=img_1,
                    img_2=img_2,
                ),
                None,
                None,
            )
        else:
            if self.weights_img_1 is None or self.weights_img_2 is None:
                self.compute_weights()

            return (
                img_1 * self.weights_img_1 + img_2 * self.weights_img_2,
                self.weights_img_1,
                self.weights_img_2,
            )


def compute_image_minimum_divergence(
    img_1: numpy.ndarray, img_2: numpy.ndarray, high_frequency_threshold, stitching_axis: int
):
    """
    Algorithm to improve treatment of high frequency.

    It split the two images into two parts: high frequency and low frequency.

    The two low frequency part will be stitched using a 'sinusoidal' / cosinus weights approach.
    When the two high frequency parts will be stitched by taking the lower divergent pixels
    """

    # split low and high frequencies
    def split_image(image: numpy.ndarray, threshold: int) -> tuple:
        """split an image to return (low_frequency, high_frequency)"""

        lowpass_filter = fourier_filters.get_lowpass_filter(
            image.shape[-2:],
            cutoff_par=threshold,
            use_rfft=True,
            data_type=image.dtype,
        )
        highpass_filter = fourier_filters.get_highpass_filter(
            image.shape[-2:],
            cutoff_par=threshold,
            use_rfft=True,
            data_type=image.dtype,
        )
        low_fre_part = local_ifftn(local_fftn(image, axes=(-2, -1)) * lowpass_filter, axes=(-2, -1)).real
        high_fre_part = local_ifftn(local_fftn(image, axes=(-2, -1)) * highpass_filter, axes=(-2, -1)).real
        return (low_fre_part, high_fre_part)

    low_freq_img_1, high_freq_img_1 = split_image(img_1, threshold=high_frequency_threshold)
    low_freq_img_2, high_freq_img_2 = split_image(img_2, threshold=high_frequency_threshold)

    # handle low frequency
    if stitching_axis == 0:
        frame_cst_size = img_1.shape[1]
        overlap_size = img_1.shape[0]
    elif stitching_axis == 1:
        frame_cst_size = img_1.shape[0]
        overlap_size = img_1.shape[1]
    else:
        raise ValueError("")

    low_freq_stitching_kernel = ImageStichOverlapKernel(
        frame_unstitched_axis_size=frame_cst_size,
        stitching_strategy=OverlapStitchingStrategy.COSINUS_WEIGHTS,
        overlap_size=overlap_size,
        stitching_axis=stitching_axis,
    )
    low_freq_stitched = low_freq_stitching_kernel.stitch(
        img_1=low_freq_img_1,
        img_2=low_freq_img_2,
        check_input=False,
    )[0]

    # handle high frequency
    mean_high_frequency = numpy.mean([high_freq_img_1, high_freq_img_2])
    assert numpy.isscalar(mean_high_frequency)
    high_freq_distance_img_1 = numpy.abs(high_freq_img_1 - mean_high_frequency)
    high_freq_distance_img_2 = numpy.abs(high_freq_img_2 - mean_high_frequency)
    high_freq_stitched = numpy.where(
        high_freq_distance_img_1 >= high_freq_distance_img_2, high_freq_distance_img_2, high_freq_distance_img_1
    )

    # merge back low and high frequencies together
    def merge_images(low_freq: numpy.ndarray, high_freq: numpy.ndarray) -> numpy.ndarray:
        """merge two part of an image. The low frequency part with the high frequency part"""
        return low_freq + high_freq

    return merge_images(low_freq_stitched, high_freq_stitched)


def compute_image_higher_signal(img_1: numpy.ndarray, img_2: numpy.ndarray):
    """
    the higher signal will pick pixel on the image having the higher signal.
    A use case is that if there is some artefacts on images which creates stripes (from scintillator artefacts for example)
    it could be removed from this method
    """
    # note: to be think about. But maybe it can be interesting to rescale img_1 and img_2
    # to ge something more coherent
    return numpy.where(img_1 >= img_2, img_1, img_2)


def check_overlaps(frames: Union[tuple, numpy.ndarray], positions: tuple, axis: int, raise_error: bool):
    """
    check over frames if there is a single overlap other juxtaposed frames (at most and at least)

    :param frames: liste of ordered / sorted frames along axis to test (from higher to lower)
    :param positions: positions of frames in 3D space as (position axis 0, position axis 1, position axis 2)
    :param axis: axis to check
    :param raise_error: if True then raise an error if two frames don't have at least and at most one overlap. Else log an error
    """
    if not isinstance(frames, (tuple, numpy.ndarray)):
        raise TypeError(f"frames is expected to be a tuple or a numpy array. Get {type(frames)} instead")
    if not isinstance(positions, tuple) and len(positions) == 3:
        raise TypeError(f"positions is expected to be a tuple of 3 elements. Get {type(positions)} instead")
    assert isinstance(axis, int), "axis is expected to be an int"
    assert isinstance(raise_error, bool), "raise_error is expected to be a bool"

    def treat_error(error_msg: str):
        if raise_error:
            raise ValueError(error_msg)
        else:
            _logger.error(raise_error)

    if axis == 0:
        axis_frame_space = 0
    elif axis == 2:
        raise NotImplementedError(f"overlap check along axis {axis_frame_space}")
    elif axis == 1:
        axis_frame_space = 1

    # convert each frame to appropriate bounding box according to the axis
    def convert_to_bb(frame: numpy.ndarray, position: tuple, axis: int):
        assert isinstance(axis, int)
        assert isinstance(position, tuple), f"position expected a tuple. Get {type(position)} instead"
        assert len(position) == 3, f"Expect to have three items for the position. Get {len(position)}"
        start_frame = position[axis] - frame.shape[axis_frame_space] // 2
        end_frame = start_frame + frame.shape[axis_frame_space]
        return BoundingBox1D(start_frame, end_frame)

    bounding_boxes = {
        convert_to_bb(frame=frame, position=position, axis=axis): position for frame, position in zip(frames, positions)
    }

    def get_frame_index(my_bb) -> str:
        bb_index = tuple(bounding_boxes.keys()).index(my_bb) + 1
        if bb_index in (1, 21, 31):
            return f"{bb_index}st"
        elif bb_index in (2, 22, 32):
            return f"{bb_index}nd"
        elif bb_index == (3, 23, 33):
            return f"{bb_index}rd"
        else:
            return f"{bb_index}th"

    # check that theres an overlap between two juxtaposed bb (or frame at the end)
    all_bounding_boxes = tuple(bounding_boxes.keys())
    bb_with_expected_overlap = [
        (bb_frame, bb_next_frame) for bb_frame, bb_next_frame in zip(all_bounding_boxes[:-1], all_bounding_boxes[1:])
    ]

    for bb_pair in bb_with_expected_overlap:
        bb_frame, bb_next_frame = bb_pair
        if bb_frame.max < bb_next_frame.min:
            treat_error(f"provided frames seems un sorted (from the higher to the lower)")
        if bb_frame.min < bb_next_frame.min:
            treat_error(
                f"Seems like {get_frame_index(bb_frame)} frame is fully overlaping with frame {get_frame_index(bb_next_frame)}"
            )
        if bb_frame.get_overlap(bb_next_frame) is None:
            treat_error(
                f"no overlap found between two juxtaposed frames - {get_frame_index(bb_frame)} and {get_frame_index(bb_next_frame)}"
            )

    # check there is no overlap between none juxtaposed bb
    def pick_all_none_juxtaposed_bb(index, my_bounding_boxes: tuple):
        """return all the bounding boxes to check for the index 'index':

        :return: (tested_bounding_box, bounding_boxes_to_test)
        """
        my_bounding_boxes = {bb_index: bb for bb_index, bb in enumerate(my_bounding_boxes)}
        bounding_boxes = dict(
            filter(
                lambda pair: pair[0] not in (index - 1, index, index + 1),
                my_bounding_boxes.items(),
            )
        )
        return my_bounding_boxes[index], bounding_boxes.values()

    bb_without_expected_overlap = [
        pick_all_none_juxtaposed_bb(index, all_bounding_boxes) for index in range(len(all_bounding_boxes))
    ]

    for bb_pair in bb_without_expected_overlap:
        bb_frame, bb_not_juxtaposed_frames = bb_pair
        for bb_not_juxtaposed_frame in bb_not_juxtaposed_frames:
            if bb_frame.get_overlap(bb_not_juxtaposed_frame) is not None:
                treat_error(
                    f"overlap found between two frames not juxtaposed - {bounding_boxes[bb_frame]} and {bounding_boxes[bb_not_juxtaposed_frame]}"
                )
